import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FileDialog;
import java.awt.Frame;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.util.Vector;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

public class SpritePanel extends JPanel implements ActionListener, ChangeListener,MouseMotionListener, ItemListener
{
	static final long serialVersionUID = 1;

	private static String BROWSE = "BROWSE";

	private JSpinner m_w,m_h,m_x,m_y;
	private SpinnerNumberModel m_wm,m_hm,m_xm,m_ym;
	private JLabel m_src,m_preview,m_path,m_coords;
	private ImageIcon m_img;
	private JComboBox m_list;
	private JCheckBox m_grid;
	
	public SpritePanel(Object[] imgs)
	{
		setLayout(new BorderLayout(8,8));
		JPanel p = new JPanel(new BorderLayout(8,8));
		m_list = new JComboBox(imgs);
		m_list.insertItemAt("",0);
		m_list.setSelectedIndex(0);
		m_list.setEditable(false);
		m_list.addItemListener(this);
		JPanel pp = new JPanel(new BorderLayout(8,8));
		pp.add(new JLabel("Images"),BorderLayout.WEST);
		pp.add(m_list,BorderLayout.CENTER);
		p.add(pp,BorderLayout.NORTH);
		pp = new JPanel(new GridLayout(1,2,8,8));
		p.add(pp,BorderLayout.CENTER);
		pp.add(m_src = new JLabel());
		m_src.addMouseMotionListener(this);
		pp.add(m_preview = new JLabel());
		Dimension d = new Dimension(256,256);
		m_src.setSize(d);
		m_src.setPreferredSize(d);
		m_src.setMinimumSize(d);
		m_preview.setSize(d);
		m_preview.setPreferredSize(d);
		m_preview.setMinimumSize(d);
		pp = new JPanel(new BorderLayout(8,8));
		pp.add(m_coords = new JLabel(),BorderLayout.NORTH);
		m_coords.setText(" ");
		pp.add(m_path = new JLabel(),BorderLayout.CENTER);
		JButton b = new JButton("Browse...");
		b.setActionCommand(BROWSE);
		b.addActionListener(this);
		pp.add(b,BorderLayout.EAST);
		p.add(pp,BorderLayout.SOUTH);
		add(p,BorderLayout.CENTER);
		p = new JPanel(new BorderLayout());
		pp = new JPanel(new GridLayout(5,1,4,4));
		pp.add(new JLabel("x"));
		pp.add(new JLabel("y"));
		pp.add(new JLabel("w"));
		pp.add(new JLabel("h"));
		pp.add(m_grid = new JCheckBox("Grid"));
		m_grid.addChangeListener(this);
		p.add(pp,BorderLayout.WEST);
		pp = new JPanel(new GridLayout(5,1,4,4));
		pp.add(m_x = new JSpinner(m_xm=new SpinnerNumberModel()));
		pp.add(m_y = new JSpinner(m_ym=new SpinnerNumberModel()));
		pp.add(m_w = new JSpinner(m_wm=new SpinnerNumberModel()));
		pp.add(m_h = new JSpinner(m_hm=new SpinnerNumberModel()));
		pp.add(new JLabel());
		m_x.addChangeListener(this);
		m_y.addChangeListener(this);
		m_w.addChangeListener(this);
		m_h.addChangeListener(this);
		p.add(pp,BorderLayout.CENTER);
		add(p,BorderLayout.SOUTH);
	}

	public SpritePanel(Object[] imgs,String p,int x,int y,int w,int h)
	{
		this(imgs);
		reset(p,x,y,w,h);
	}

	public Rectangle getRectangle()
	{
		return new Rectangle(
					m_xm.getNumber().intValue(),
					m_ym.getNumber().intValue(),
					m_wm.getNumber().intValue(),
					m_hm.getNumber().intValue());
	}

	public String getPath()
	{
		return m_path.getText();
	}

	public static Image extractImg(Image img,Rectangle r)
	{
		BufferedImage bimg = new BufferedImage(r.width,r.height,BufferedImage.TYPE_INT_ARGB);
		bimg.getGraphics().drawImage(img,0,0,r.width,r.height,r.x,r.y,r.x+r.width,r.y+r.height,null);
		return bimg;
	}

	public Image getImage()
	{
		if (m_img==null || m_img.getImage()==null)
			return null;
		if (m_grid.isSelected())
			return m_img.getImage();
		return extractImg(m_img.getImage(),getRectangle());
	}
	
	public boolean isGrid()
	{
		return m_grid.isSelected();
	}

	public Vector<Object> getImages()
	{
		if (m_img==null || m_img.getImage()==null)
			return null;
		Vector<Object> v = new Vector<Object>();
		if (m_grid.isSelected())
		{
			Rectangle r = new Rectangle();
			r.width  = m_wm.getNumber().intValue();
			r.height = m_hm.getNumber().intValue();
			for (r.y=0;r.y<m_img.getIconHeight();r.y+=r.height)
			{
				for (r.x=0;r.x<m_img.getIconWidth();r.x+=r.width)
				{
					v.addElement(extractImg(m_img.getImage(),r));
					v.addElement(new Rectangle(r));
				}
			}
		}
		else
			v.addElement(getImage());
		return v;
	}

	public void actionPerformed(ActionEvent e)
	{
		File f = TrackMain.selectFile(new Frame(),"Select sprite source",FileDialog.LOAD);
		if (f!=null)
		{
			m_list.setSelectedIndex(0);
			reset(f.getAbsolutePath(),0,0,0,0);
		}
	}

	private void reset(String p,int x,int y,int w,int h)
	{
		m_img = new ImageIcon(p); 
		m_src.setIcon(m_img);
		m_path.setText(p);
		m_x.setModel(m_xm = new SpinnerNumberModel(x,0,m_img.getIconWidth(),1));
		m_y.setModel(m_ym = new SpinnerNumberModel(y,0,m_img.getIconHeight(),1));
		m_w.setModel(m_wm = new SpinnerNumberModel(w==0?m_img.getIconWidth():w,1,m_img.getIconWidth(),1));
		m_h.setModel(m_hm = new SpinnerNumberModel(h==0?m_img.getIconHeight():h,1,m_img.getIconHeight(),1));
		m_preview.setIcon(new ImageIcon(getImage()));
		m_grid.setSelected(false);
	}

	public void stateChanged(ChangeEvent e)
	{
		if (e.getSource()==m_grid)
		{
			boolean b = m_grid.isSelected();
			m_x.setEnabled(!b);
			m_y.setEnabled(!b);
		}
		m_preview.setIcon(new ImageIcon(getImage()));
	}

	public void mouseDragged(MouseEvent e)
	{
	}

	public void mouseMoved(MouseEvent e)
	{
		m_coords.setText("X="+e.getX()+" Y="+e.getY());
	}
	
	public String toString()
	{
		return getName();
	}

	public void itemStateChanged(ItemEvent e)
	{
		if (m_list.getSelectedIndex()>0)
			reset(m_list.getSelectedItem().toString(),0,0,0,0);
	}
}
