package j3d;

import java.awt.*;

public class ColoredTriangle3D extends Triangle3D {

  protected Color color;
  protected int red,green,blue;
  private ColoredTriangle3D cpy = null;
  private boolean overload;

  public ColoredTriangle3D(Object3D obj, int a, int b, int c, Color color) {
    this(obj,null,a,b,c,color);
  }

  public ColoredTriangle3D(Object3D obj, Triangle3D origin, int a, int b, int c, Color color) {
    super(obj,origin,a,b,c);
    overload = false;
    setColor(color);
  }

  public boolean isOverloaded()
  {
  	return overload;
  }
  
  public void setColor(Color color) {
    this.color = color;
    if (color!=null) {
       red=color.getRed();
       green=color.getGreen();
       blue=color.getBlue();
    }
  }
    
  public void overloadColor(Color color)
  {
  	setColor(color);
    overload = true;
  }

  public Color getColor() {
    return this.color;
  }

  public Triangle3D copy(Object3D obj,int d) {
    if (cpy==null) cpy = new ColoredTriangle3D(obj,origin==null?this:origin,a+d,b+d,c+d,color);
    else
    {
       cpy.a = a + d;
       cpy.b = b + d;
       cpy.c = c + d;
       cpy.color = color;
       cpy.red = red;
       cpy.green = green;
       cpy.blue =blue;
    }
    cpy.alpha = alpha;
    return cpy;
  }

  public Triangle3D copy(Object3D obj)
  {
    return new ColoredTriangle3D(obj,null,a,b,c,color);
  }

  // Gouraud Shading
  public void paint(int[] img,float[] zbuf,Triangle3D[] tmap,Rectangle r,int[][] c_lights,float[] f_lights) {
    initPoints();
    int A,B,C,xA,xB,xC,yA,yB,yC;
    float zA,zB,zC;

    if (py[0]<py[1]) {
       if (py[1]<py[2]) {
          A = a; B = b; C = c;
          xA = px[0]; xB = px[1]; xC = px[2];
          yA = py[0]; yB = py[1]; yC = py[2];
          zA = pz[0]; zB = pz[1]; zC = pz[2];
       } else {
          if (py[0]<py[2]) {
             A = a; B = c; C = b;
             xA = px[0]; xB = px[2]; xC = px[1];
             yA = py[0]; yB = py[2]; yC = py[1];
             zA = pz[0]; zB = pz[2]; zC = pz[1];
          } else {
             A = c; B = a; C = b;
             xA = px[2]; xB = px[0]; xC = px[1];
             yA = py[2]; yB = py[0]; yC = py[1];
             zA = pz[2]; zB = pz[0]; zC = pz[1];
          }
       }
    } else {
       if (py[0]<py[2]) {
          A = b; B = a; C = c;
          xA = px[1]; xB = px[0]; xC = px[2];
          yA = py[1]; yB = py[0]; yC = py[2];
          zA = pz[1]; zB = pz[0]; zC = pz[2];
       } else {
          if (py[1]<py[2]) {
             A = b; B = c; C = a;
             xA = px[1]; xB = px[2]; xC = px[0];
             yA = py[1]; yB = py[2]; yC = py[0];
             zA = pz[1]; zB = pz[2]; zC = pz[0];
          } else {
             A = c; B = b; C = a;
             xA = px[2]; xB = px[1]; xC = px[0];
             yA = py[2]; yB = py[1]; yC = py[0];
             zA = pz[2]; zB = pz[1]; zC = pz[0];
          }
       }
    }

    if ((yA+r.y>r.height)||(yC+r.y<0)||(Math.max(Math.max(xA,xB),xC)+r.x<0)||(Math.min(Math.min(xA,xB),xC)+r.x>r.width)) return;

    int _wx = r.width - r.x;
    int _hy = r.height - r.y;
    if ((xA>=-r.x&&xA<_wx)&&(xB>=-r.x&&xB<_wx)&&(xC>=-r.x&&xC<_wx)&&
        (yA>=-r.y&&yA<_hy)&&(yB>=-r.y&&yB<_hy)&&(yC>=-r.y&&yC<_hy))
    {
       int _dA = (r.x+xA)+(r.y+yA)*r.width;
       Triangle3D _t = tmap[_dA];
       if (_t!=null)
       {
          int _dB = (r.x+xB)+(r.y+yB)*r.width,
              _dC = (r.x+xC)+(r.y+yC)*r.width;
          if((_t==tmap[_dB])&&(_t==tmap[_dC])&&(zA<=zbuf[_dA])&&(zB<=zbuf[_dB])&&(zC<=zbuf[_dC]))
             return;
       }
    }

    int ra = computeColor(c_lights[A][0],f_lights[A],red);
    int ga = computeColor(c_lights[A][1],f_lights[A],green);
    int ba = computeColor(c_lights[A][2],f_lights[A],blue);
    int rb = computeColor(c_lights[B][0],f_lights[B],red);
    int gb = computeColor(c_lights[B][1],f_lights[B],green);
    int bb = computeColor(c_lights[B][2],f_lights[B],blue);
    int rc = computeColor(c_lights[C][0],f_lights[C],red);
    int gc = computeColor(c_lights[C][1],f_lights[C],green);
    int bc = computeColor(c_lights[C][2],f_lights[C],blue);

    AB = line(xA,yA,xB,yB,AB);
    AC = line(xA,yA,xC,yC,AC);
    BC = line(xB,yB,xC,yC,BC);

    int ia,x2,k;
    int xl,xr=0,rl,gl,bl,rr=0,gr=0,br=0;
    int r2,g2,b2;
    int yBA = yB - yA, yCA = yC - yA, yCB = yC - yB;
    float m,zl,zr,z2;
    int rwx = _wx - 1;
    int maxy = _hy - 1;
    maxy = (maxy<yC?maxy:yC)-1;
    int miny = yA>-r.y?yA:-r.y;
    int di = r.x+(miny+r.y)*r.width;

    for (int i=miny;i<=maxy;i++) {
        if (i<=yB) {
           xl = AB[ia=i-yA];
           x2 = AC[ia];
           m = (yBA == 0 ? 1f : ((float)ia)/yBA);
           rl = pointAt(ra,rb,m);
           gl = pointAt(ga,gb,m);
           bl = pointAt(ba,bb,m);
           zl = pointAt(zA,zB,m);
        } else {
           xl = BC[i-yB];
           x2 = AC[ia=i-yA];
           m = (yCB == 0 ? 1f : ((float)(i-yB))/yCB);
           rl = pointAt(rb,rc,m);
           gl = pointAt(gb,gc,m);
           bl = pointAt(bb,bc,m);
           zl = pointAt(zB,zC,m);
        }
        if (xl==x2) {
           if ((xl>1-r.x)&&(xl<rwx)) {
              if (zl>zbuf[k=xl+di]) {
                 img[k] = (alpha<255 ? computeAlpha(img[k],rl,gl,bl) : 0xff000000|(rl<<16)|(gl<<8)|bl);
                 zbuf[k] = zl;
                 tmap[k] = this;
              }
           }
        } else {
           m = (yCA == 0 ? 1f : ((float)ia)/yCA);
           r2 = pointAt(ra,rc,m);
           g2 = pointAt(ga,gc,m);
           b2 = pointAt(ba,bc,m);
           z2 = pointAt(zA,zC,m);
           if (xl<x2) {
              xr = x2;
              zr = z2;
              rr = r2; gr = g2; br = b2;
           } else {
              xr = xl;
              zr = zl;
              rr = rl; gr = gl; br = bl;
              xl = x2;
              zl = z2;
              rl = r2; gl = g2; bl = b2;
           }
           int minx = xl>1-r.x?xl:1-r.x;
           int dx = xr-xl,maxx = (xr<rwx?xr:rwx)-1, begx = minx - xl;
           k = minx-1+di;
           float dr = ((float)(rr-rl))/dx,
                 dg = ((float)(gr-gl))/dx,
                 db = ((float)(br-bl))/dx,
                 dz = (zr-zl)/dx,
                 _r = rl+dr*begx,
                 _g = gl+dg*begx,
                 _b = bl+db*begx,
                 zp = zl+dz*begx;
           if (alpha<255) {
              for (int j=minx;j<=maxx;j++) {
                  if (zp>zbuf[++k]) {
                     img[k] = computeAlpha(img[k],(int)_r,(int)_g,(int)_b);
                     zbuf[k] = zp;
                     tmap[k] = this;
                  }
                  _r += dr;
                  _g += dg;
                  _b += db;
                  zp += dz;
              }
           } else {
              for (int j=minx;j<=maxx;j++) {
                  if (zp>zbuf[++k]) {
                     img[k] = 0xff000000|(((int)_r)<<16)|(((int)_g)<<8)|(int)_b;
                     zbuf[k] = zp;
                     tmap[k] = this;
                  }
                  _r += dr;
                  _g += dg;
                  _b += db;
                  zp += dz;
              }
           }
        }
        di += r.width;
    }
  }
}
