package j3d;

import java.util.Vector;
import java.awt.Point;

public class DefaultScene3D extends AbstractTransformableObject3D implements Scene3D {

  protected Vector<Object> objs = new Vector<Object>();
  protected Vector<Position3D> opos = new Vector<Position3D>();
  private String name;

  public Point3D[] vertices = null;
  public Triangle3D[] triangles = null;
  public Vector3D[] vnormals = null;
  public Vector3D[] tnormals = null;

  public DefaultScene3D() {
    this("noname");
  }
  
  public DefaultScene3D(String name) {
    this.name = name;
  }

  public String toString()
  {
    return name;
  }

  public void setElements(Point3D[] v,Triangle3D[] t,Vector3D[] vnorm,Vector3D[] tnorm) {
    vertices = v;
    triangles = t;
    vnormals = vnorm;
    tnormals = tnorm;
  }

  public Object[] getObjects() {
    Object[] ret = new Object[objs.size()];
    objs.copyInto(ret);
    return ret;
  }

  public Position3D getPosition(Object obj) {
    int i = objs.indexOf(obj);
    return i < 0 ? null : (Position3D)opos.elementAt(i);
  }

  public void addObject(Object obj,Position3D pos)
  {
    if (objs.contains(obj))
    	return;
    objs.addElement(obj);
    opos.addElement(pos);
  }

  public void setObjectAt(Object obj, Position3D pos, int idx) {
  	objs.setElementAt(obj,idx);
  	opos.setElementAt(pos,idx);
  }

  public void removeObject(Object obj) {
    int i = objs.indexOf(obj);
    if (i<0) {
       for (i=0;i<objs.size();i++) {
           Object o = objs.elementAt(i);
           if (o instanceof Scene3D) ((Scene3D)o).removeObject(obj);
       }
       return;
    }
    objs.removeElementAt(i);
    opos.removeElementAt(i);
  }

  public void transform(Object3D o,Matrix3D cur,Position3D pos,Point3D[] v,Triangle3D[] t,Vector3D[] vnorm,Vector3D[] tnorm,Point dVT) {
    Matrix3D m = pos.getTransMatrix();
    m = m.multiply(cur);
    int osize = objs.size();
    for (int i=0;i<osize;i++)
    {
    	Object obj = objs.elementAt(i);
    	if (obj instanceof Object3D)
    		((Object3D)obj).transform(o,m,opos.elementAt(i),v,t,vnorm,tnorm,dVT);
    	else
    		((Sprite3D)obj).transform(m,opos.elementAt(i));
    }
  }

  public Point3D[] getVertices() {
    return vertices;
  }

  public Triangle3D[] getTriangles() {
    return triangles;
  }

  public Vector3D[] getVNormals() {
    return vnormals;
  }

  public Vector3D[] getTNormals() {
    return tnormals;
  }

  public Point3D[] getBoundingBox()
  {
	Point3D[] box = new Point3D[2];
  	box[0] = new Point3D();
  	box[1] = new Point3D();
  	for (int i=0;i<objs.size();i++)
  	{
  		Object3D o = (Object3D)objs.elementAt(i);
  		Point3D[] bb = o.getBoundingBox();
  		if (i==0 || bb[0].x<box[0].x)
  			box[0].x = bb[0].x;
  		if (i==0 || bb[0].y<box[0].y)
  			box[0].y = bb[0].y;
  		if (i==0 || bb[0].z<box[0].z)
  			box[0].z = bb[0].z;
  		if (i==0 || bb[1].x>box[1].x)
  			box[1].x = bb[1].x;
  		if (i==0 || bb[1].y>box[1].y)
  			box[1].y = bb[1].y;
  		if (i==0 || bb[1].z>box[1].z)
  			box[1].z = bb[1].z;
  	}
  	return box;
  }

  public void center()
  {
	Point3D[] bb = getBoundingBox();
	float dx = (bb[0].x + bb[1].x)/2;
	float dy = (bb[0].y + bb[1].y)/2;
	float dz = (bb[0].z + bb[1].z)/2;
  	for (int i=0;i<objs.size();i++)
  	{
  		Object3D o = (Object3D)objs.elementAt(i);
  		Point3D[] v = o.getVertices();
  		for (int j=0;j<v.length;j++)
  		{
  			v[j].x-=dx;
  			v[j].y-=dy;
  			v[j].z-=dz;
  		}
	}
  }
  
  public int getVCount()
  {
    int n = 0;
  	for (int i=0;i<objs.size();i++)
  	{
  		Object3D o = (Object3D)objs.elementAt(i);
  		n += o.getVertices().length;
    }
  	return n;
  }
  
  public int getTCount()
  {
    int n = 0;
  	for (int i=0;i<objs.size();i++)
  	{
  		Object3D o = (Object3D)objs.elementAt(i);
  		n += o.getTriangles().length;
    }
  	return n;
  }
}
