package j3d;

import java.awt.Point;
import java.awt.Rectangle;
import java.util.ArrayList;

public class DefaultView3D implements View3D {
	
	public static final int MAX_TRIANGLES = 50000;
	public static final int MAX_VERTICES = MAX_TRIANGLES*3;
	
	// Triangles
	private static Triangle3D[] t = new Triangle3D[MAX_TRIANGLES];
	// Vertices
	private static Point3D[] v = new Point3D[MAX_VERTICES];
	
	// Normals
	private static Point3D[] vnorm = new Point3D[MAX_VERTICES];
	private static Point3D[] tnorm = new Point3D[MAX_TRIANGLES];
	
	private static Triangle3D[] nt = new Triangle3D[MAX_TRIANGLES];
	private static Triangle3D[] at = new Triangle3D[MAX_TRIANGLES]; // alpha triangles
	
	// Sort
	private static float[] z = new float[MAX_TRIANGLES];
	private static int[] a = new int[MAX_TRIANGLES];
	private static int[] b = new int[MAX_TRIANGLES];
	private static int[] c = new int[MAX_TRIANGLES];
	
	// lights
	private static float[][] dlights = null;
	private static float[] f_lights = null;
	private static int[][] c_lights = null;
	
	// Visibility
	private static boolean[] vis = new boolean[MAX_VERTICES];
	
	public DefaultView3D() {
		for (int i=0;i<v.length;i++) {
			v[i] = new Point3D();
			vnorm[i] = new Point3D();
		}
		for (int i=0;i<t.length;i++)
			tnorm[i] = new Point3D();
	}
	
	public void render(int[] img,float[] zbuf,Triangle3D[] tmap,Rectangle rect, Scene3D scene, Position3D camera,float proj,float maxz,Light3D[] lights) {
		Point dVT = new Point(0,0);
		scene.transform(scene,new Matrix3D(Matrix3D.IDENT),camera,v,t,vnorm,tnorm,dVT);
		scene.setElements(v,t,vnorm,tnorm);
		int vl = dVT.x,tl = dVT.y,ntt = 0, nat = 0;
		float mproj = -((float)Math.sqrt(rect.width*rect.height)*proj)/20,minz;
		java.util.Arrays.fill(zbuf,Float.NEGATIVE_INFINITY);
		java.util.Arrays.fill(img,0xff7f7f7f);
		java.util.Arrays.fill(tmap,null);
		java.util.Arrays.fill(vis,0,vl,false);

		for (int i=0;i<tl;i++)
		{
			Triangle3D tt = t[i];
			minz=tt.minZ();
			if ((minz<maxz)&&(minz>mproj))
			{
				if (tt.alpha<255)
				{
					z[nat] = minz;
					a[nat] = nat;
					at[nat++] = tt;
					vis[tt.a] = vis[tt.b] = vis[tt.c] = true;
				}
				else
					if (tnorm[i].z>=0)
					{
						nt[ntt++] = tt;
						vis[tt.a] = vis[tt.b] = vis[tt.c] = true;
					}
			}
		}

		if (dlights==null||dlights[0].length!=lights.length)
		{
			c_lights = new int[MAX_VERTICES][3];
			dlights = new float[MAX_VERTICES][lights.length];
			f_lights = new float[MAX_VERTICES];
		}

		/* lights positions and intensity */
		int ln = lights.length;
		Point3D vc = new Point3D(), p = new Point3D();
		Matrix3D m = camera.getTransMatrix();
		for (int j=0;j<ln;j++) {
			Light3D lj = lights[j];
			m.multiply(lj.loc,p);
			/* Lights intensity */
			for (int i=0;i<vl;i++) {
				if (vis[i]) {
					vc.x = v[i].x-p.x; vc.y = v[i].y-p.y; vc.z = v[i].z-p.z;
					float l = vc.length();
					vc.normalize(l);
					vc.add(vnorm[i]);
					float f  = (lj.intensity==0?2f:l<lj.intensity?1f:l/lj.intensity);
					dlights[i][j] = (1-vc.length()/2)/f;
				}
			}
		}

		/* Lights on each vertex */
		for (int i=0;i<vl;i++) {
			if (vis[i]) {
				v[i].x *= proj/v[i].z;
				v[i].y *= proj/v[i].z;
				float r = 0f, g = 0f, b = 0f;
				for (int j=0;j<ln;j++) {
					Light3D lj = lights[j];
					float dl = dlights[i][j];
					r += lj.red*dl;
					g += lj.green*dl;
					b += lj.blue*dl;
				}
				c_lights[i][0] = (int)(r = (r>255f ? 255f : r));
				c_lights[i][1] = (int)(g = (g>255f ? 255f : g));
				c_lights[i][2] = (int)(b = (b>255f ? 255f : b));
				f_lights[i] = (r+g+b)/765f; // 255*3
			}
		}
		
		/* Paint all */
		for (int i=0;i<ntt;i++)
			nt[i].paint(img,zbuf,tmap,rect,c_lights,f_lights);
		ArrayList<Object[]> list = new ArrayList<Object[]>();
		list.add(scene.getObjects());
		for (int i=0;i<list.size();i++)
		{
			Object[] objs = list.get(i);
			for (int j=0;j<objs.length;j++)
			{
				if (objs[j] instanceof Scene3D)
					list.add(((Scene3D)objs[j]).getObjects());
				else if (objs[j] instanceof Sprite3D)
				{
					Sprite3D spr = (Sprite3D)objs[j]; 
					if ((spr.center.z<maxz)&&(spr.center.z>mproj))
						spr.paint(img,zbuf,rect,proj);
				}
			}
		}
		if (nat>0)
		{
			qsort(a,b,c,z,0,nat);
			for (int i=0;i<nat;i++)
				at[c[i]].paint(img,zbuf,tmap,rect,c_lights,f_lights);
		}
	}
	
	public static void qsort(int[] a,int[] b,int[] c,float[] v,int start,int end)
	{
		int aa,i,j=start,k=end;
		float va;
		if (end<=start)
			return;
		aa = a[start];
		if (start+1==end)
		{
			c[start] = aa;
			return;
		}
		va = v[aa];
		for (i=start+1;i<end;i++)
			b[(v[a[i]]<va)?j++:--k] = a[i];
		c[j] = aa;
		qsort(b,a,c,v,start,j);
		qsort(b,a,c,v,j+1,end);
	}
}
