package j3d;

public class Matrix3D {

  public static final int VAL    = 0;
  public static final int ROTX   = 1;
  public static final int ROTY   = 2;
  public static final int ROTZ   = 3;
  public static final int ROTXYZ = 4;
  public static final int SCALE  = 5;
  public static final int TRANS  = 6;
  public static final int IDENT  = 7;
  public static final int PROJ   = 8;

  public float[] values;

  public Matrix3D()					{ this(VAL,0); }
  public Matrix3D(int type)			{ this(type,0,0,0); }
  public Matrix3D(int type,float v) { this(type,v,v,v); }
  public Matrix3D(float[] v)		{ values = v; }

  public Matrix3D(int type,float v0,float v1,float v2) {
    switch(type) {
      case VAL:
        values = new float[] {v0,v0,v0,v0,v0,v0,v0,v0,v0,v0,v0,v0,v0,v0,v0,v0};
        break;
      case ROTX:
        values = new float[] {1,0,0,0,0,(float)Math.cos(v0),(float)Math.sin(v0),0,0,-(float)Math.sin(v0),(float)Math.cos(v0),0,0,0,0,1};
        break;
      case ROTY:
        values = new float[] {(float)Math.cos(v0),0,-(float)Math.sin(v0),0,0,1,0,0,(float)Math.sin(v0),0,(float)Math.cos(v0),0,0,0,0,1};
        break;
      case ROTZ:
        values = new float[] {(float)Math.cos(v0),(float)Math.sin(v0),0,0,-(float)Math.sin(v0),(float)Math.cos(v0),0,0,0,0,1,0,0,0,0,1};
        break;
      case ROTXYZ:
        float cx = (float)Math.cos(v0), cy = (float)Math.cos(v1), cz = (float)Math.cos(v2);
        float sx = (float)Math.sin(v0), sy = (float)Math.sin(v1), sz = (float)Math.sin(v2);

        values = new float[] {cy*cz,          cy*sz,         -sy,    0,
                              sx*sy*cz-cx*sz, sx*sy*sz+cx*cz, sx*cy, 0,
                              cx*sy*cz+sx*sz, cx*sy*sz-sx*cz, cx*cy, 0,
                              0,              0,              0,     1};
/*
        values = new float[] {cz*cy-sz*sx*sy, -sz*cx, cz*sy+sz*sx*cy, 0,
                              sz*cy+cz*sx*sy, cz*cx , sz*sy-cz*sx*cy, 0,
                              -cz*sy        , sx    , cx*cy         , 0,
                              0             , 0     , 0             , 1};
*/
        break;
      case SCALE:
        values = new float[] {v0,0,0,0,0,v1,0,0,0,0,v2,0,0,0,0,1};
        break;
      case TRANS:
        values = new float[] {1,0,0,0,0,1,0,0,0,0,1,0,v0,v1,v2,1};
        break;
      case IDENT:
        values = new float[] {1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1};
        break;
      case PROJ:
        values = new float[] {1,0,0,0,0,1,0,0,0,0,1,v0,0,0,0,1};
        break;
    }
  }

  public Matrix3D multiply(Matrix3D m) {
    Matrix3D q = new Matrix3D();
    for(int i=0;i<4;i++)
        for(int j=0;j<4;j++)
            for(int k=0;k<4;k++)
                q.values[i*4+j] += m.values[k*4+j]*this.values[i*4+k];
    return q;
  }

  public Point3D multiply(Point3D v) {
    Point3D res = new Point3D();
    res.x = v.x * values[0] + v.y * values[4] + v.z *values[8]  + values[12];
    res.y = v.x * values[1] + v.y * values[5] + v.z *values[9]  + values[13];
    res.z = v.x * values[2] + v.y * values[6] + v.z *values[10] + values[14];
    return res;
  }

  public void multiply(Point3D v,Point3D res) {
    float x = v.x * values[0] + v.y * values[4] + v.z *values[8]  + values[12];
    float y = v.x * values[1] + v.y * values[5] + v.z *values[9]  + values[13];
    res.z = v.x * values[2] + v.y * values[6] + v.z *values[10] + values[14];
    res.x = x;
    res.y = y;
  }

  public void multiply(Vector3D v,Vector3D res) {
    float x = v.x * values[0] + v.y * values[4] + v.z *values[8];
    float y = v.x * values[1] + v.y * values[5] + v.z *values[9];
    res.z = v.x * values[2] + v.y * values[6] + v.z *values[10];
    res.x = x;
    res.y = y;
  }

  public void print() {
    for (int i=0;i<16;i++) {
        System.err.print(values[i]+" ");
        if (i%4==3) System.err.println();
    }
    System.err.println();
  }
}
