package j3d;

import java.awt.*;

public abstract class Triangle3D implements Comparable {

  public Object3D obj;
  public Triangle3D origin;
  public int a,b,c,alpha;
  public static int px[]   = new int[3];
  public static int py[]   = new int[3];
  public static float pz[] = new float[3];
  public static int AB[] = new int[1024];
  public static int AC[] = new int[1024];
  public static int BC[] = new int[1024];

  public Triangle3D(Object3D obj,Triangle3D origin,int a, int b, int c) {
    this.obj = obj;
    this.origin = origin;
    this.a = a;
    this.b = b;
    this.c = c;
    this.alpha = 255;
  }

  public abstract Triangle3D copy(Object3D obj,int d);

  protected void initPoints() {
    Point3D[] v = obj.getVertices();
    px[0]=(int)v[a].x; py[0]=(int)v[a].y; pz[0]=(int)v[a].z;
    px[1]=(int)v[b].x; py[1]=(int)v[b].y; pz[1]=(int)v[b].z;
    px[2]=(int)v[c].x; py[2]=(int)v[c].y; pz[2]=(int)v[c].z;
  }

  public float getZ() {
    Point3D[] v = obj.getVertices();
    return (v[a].z+v[b].z+v[c].z)/3;
  }

  public float minZ() {
    Point3D[] v = obj.getVertices();
    return Math.min(Math.min(v[a].z,v[b].z),v[c].z);
  }

  public void getNormal(Point3D vn) {
    Point3D[] v = obj.getVertices();
    vn.x = ((v[c].y-v[b].y)*(v[b].z-v[a].z))-((v[c].z-v[b].z)*(v[b].y-v[a].y));
    vn.y = ((v[c].z-v[b].z)*(v[b].x-v[a].x))-((v[c].x-v[b].x)*(v[b].z-v[a].z));
    vn.z = ((v[c].x-v[b].x)*(v[b].y-v[a].y))-((v[c].y-v[b].y)*(v[b].x-v[a].x));
  }

  public Point3D getCenter() {
    Point3D[] v = obj.getVertices();
    return new Point3D((v[a].x+v[b].x+v[c].x)/3,(v[a].y+v[b].y+v[c].y)/3,(v[a].z+v[b].z+v[c].z)/3);
  }

  public int compareTo(Object o) {
    return (int)(getZ()-((Triangle3D)o).getZ());
  }

  public int[] line(int xA,int yA,int xB,int yB,int[] ret) {
    int dy = yB-yA;
    if (dy+1>ret.length) ret = new int[dy+1];
    int dx = xB-xA;
    if (dx==0) java.util.Arrays.fill(ret,0,dy+1,xA);
    else {
       ret[0] = xA;
       ret[dy] = xB;
       float aAB = ((float)dy)/dx, bAB = yB-aAB*xB, c = (yA-bAB)/aAB, iaAB = 1/aAB;
       for (int i=1;i<dy;i++) ret[i] = (int)(c+=iaAB);
    }
    return ret;
  }

  public static int pointAt(int u,int v,float coef) {
    return (int)(v*coef+u*(1-coef));
  }

  public static float pointAt(float u,float v,float coef) {
    return v*coef+u*(1-coef);
  }

  public static int computeColor(int c_lights,float f_lights,int col) {
    int ret = (int)(col*f_lights+c_lights);
    return ret > 255 ? 255 : ret;
  }

  public abstract void paint(int[] img,float[] zbuf,Triangle3D[] tmap,Rectangle r,int[][] c_lights,float[] f_lights);

  public int computeAlpha(int pi,int r,int g,int b) {
     int ri = (pi>>16)&0xff, gi = (pi>>8)&0xff, bi = pi&0xff, beta = 255-alpha;
     ri = ((r*alpha)+(ri*beta))/255;
     gi = ((g*alpha)+(gi*beta))/255;
     bi = ((b*alpha)+(bi*beta))/255;
     return 0xff000000|(ri<<16)|(gi<<8)|bi;
  }

  public int getIndex()
  {
    Triangle3D[] t = origin.obj.getTriangles();
    for (int i=0;i<t.length;i++)
        if (t[i]==origin)
           return i;
    return -1;
  }

  public boolean isNear(Triangle3D t)
  {
    return ((t.a==a || t.a==b || t.a==c)||(t.b==a || t.b==b || t.b==c)||(t.c==a || t.c==b || t.c==c));
  }

  public abstract Triangle3D copy(Object3D obj);
}
