package j3d.basic;

import java.util.Random;

public abstract class FractalMap extends Map {

  private static final int MY = 64;
  private static final int DY = 128;
  private static final int SEED = 2;
  private transient Random rand;

  public int m_seed;
  public int m_my;
  public int m_dy;
  public int m_smooth;
  public boolean m_rot;
  private float[][] m_terra;

  public FractalMap(int size,int smooth) {
    this(size,size,smooth,MY,DY,SEED,false);
  }

  public FractalMap(int w,int h,int smooth,int my,int dy) {
    this(w,h,smooth,MY,DY,SEED,false);
  }

  public FractalMap(int w,int h,int smooth,int my,int dy,int seed,boolean rot) {
	super(w,h);
	m_smooth = smooth;
	m_my     = my;
    m_dy     = dy;
    m_rot    = rot;
	m_seed   = seed;
	init();
  }

  public void initZ() {
    int w1 = m_w+1;
    int h1 = m_h+1;

    m_terra = new float[w1][h1];
    rand = new Random(m_seed);
    m_terra[0][0] = rand.nextFloat()*m_my;
    if (m_rot) {
       m_terra[0][m_h] = m_terra[m_w][0] = m_terra[m_w][m_h] = m_terra[0][0];
    } else {
       m_terra[0][m_h] = rand.nextFloat()*m_my;
       m_terra[m_w][0] = rand.nextFloat()*m_my;
       m_terra[m_w][m_h] = rand.nextFloat()*m_my;
    }
    computeFractal(0,0,m_w,m_h,m_dy,m_rot,m_w,m_h);
    for (int i=0;i<m_smooth;i++) lissage(m_w,m_h,m_rot);
  }

  public float getZ(int i,int j) {
    return m_terra[i][j];
  }

  public void computeFractal(int x1,int y1,int x2,int y2,int dy,boolean rot,int w,int h) {
    if (x2-x1<2 && y2-y1<2) return;
    int a=(x1+x2)/2, b=(y1+y2)/2;
    m_terra[x1][b]=(m_terra[x1][y1]+m_terra[x1][y2])/2+rand.nextFloat()*dy-rand.nextFloat()*dy;
    m_terra[a][y1]=(m_terra[x1][y1]+m_terra[x2][y1])/2+rand.nextFloat()*dy-rand.nextFloat()*dy;
    m_terra[a][b]=(m_terra[x1][y1]+m_terra[x1][y2]+m_terra[x2][y1]+m_terra[x2][y2])/4+rand.nextFloat()*dy-rand.nextFloat()*dy;
    if (rot && x2 == w)
       m_terra[x2][b]=m_terra[0][b];
    else
       m_terra[x2][b]=(m_terra[x2][y1]+m_terra[x2][y2])/2+rand.nextFloat()*dy-rand.nextFloat()*dy;
    if (rot && y2 == w)
       m_terra[a][y2]=m_terra[a][0];
    else
       m_terra[a][y2]=(m_terra[x1][y2]+m_terra[x2][y2])/2+rand.nextFloat()*dy-rand.nextFloat()*dy;
    computeFractal(x1,y1,a,b,dy,rot,w,h);
    computeFractal(x1,b,a,y2,dy,rot,w,h);
    computeFractal(a,y1,x2,b,dy,rot,w,h);
    computeFractal(a,b,x2,y2,dy,rot,w,h);
  }

  public void lissage(int w,int h,boolean rot) {
     float[][] lterra = new float[w+1][h+1];
     int a,b,c,d;
     for (int i=0;i<=w;i++) {
         for (int j=0;j<=h;j++) {
             a = i==0 ? w-1 : i-1;
             b = i==w ? 0   : i+1;
             c = j==0 ? h-1 : j-1;
             d = j==h ? 0   : j+1;
             if (rot && i==w)
                lterra[i][j]=lterra[0][j];
             else if (rot && j==h)
                lterra[i][j]=lterra[i][0];
             else
                lterra[i][j]=(m_terra[a][c]+m_terra[i][c]+m_terra[b][c]+
                              m_terra[a][j]+m_terra[i][j]+m_terra[b][j]+
                              m_terra[a][d]+m_terra[i][d]+m_terra[b][d])/9;
         }
     }
     for (int i=0;i<=w;i++)
         for (int j=0;j<=h;j++) m_terra[i][j] = lterra[i][j];
  }
}
