package j3d.basic;

import java.io.LineNumberReader;
import java.io.PrintWriter;

import j3d.Point3D;

public abstract class Track extends BasicObject3D
{
	public Point3D[] traj;
	public float[] roty;
	public Point3D[] left;
	public Point3D[] right;
	public Point3D[] farleft;
	public Point3D[] farright;
	public int[][] data;

	public Track(int div,LineNumberReader lnr) throws Exception
	{
		this(div);
		if (lnr!=null)
			loadTrack(lnr,true);
	}
	
	public Track(int div)
	{
		reset(div);
	}
	
	public void reset(int div)
	{
		vertices = new Point3D[div*8];
		traj = new Point3D[div];
		roty = new float[div];
		left = new Point3D[div];
		right = new Point3D[div];
		farleft = new Point3D[div];
		farright = new Point3D[div];
		data = new int[div][7];
		alloc(div*14);
		int nbv=0,nbt=0;
		double a = 0,da = (2*Math.PI)/div;
		double r = Math.pow(3,div/128);
		double rx0 = 1000*r,rz0 = 1000*r;
		double rx1 = rx0+200, rz1 = rz0+200;
		double ay = 0,day = (8*Math.PI)/div;
		double az = 0,daz = (4*Math.PI)/div;
		for (int i=0;i<div;i++)
		{
			double x0 = Math.cos(a)*rx0,z0 = Math.sin(a)*rz0;
			double x1 = Math.cos(a)*rx1,z1 = Math.sin(a)*rz1;
			double dx = (x1-x0)/5;
			double dz = (z1-z0)/5;
			int y[] = {30,30,0,0,0,0,30,30};
			int x[] = {-4,0,1,2,3,4,5,9};
			double yi = (Math.cos(ay)+Math.sin(ay))*20;
			double dd = (Math.cos(az)+Math.sin(2*az))*100;
			for (int j=0;j<8;j++)
				vertices[nbv++] = new Point3D((float)(dd+x0+x[j]*dx),(float)(y[j]+yi),(float)(dd+z0+x[j]*dz));
			left[i] = new Point3D((float)(dd+x0-dx),(float)(y[0]+yi),(float)(dd+z0-dz));
			right[i] = new Point3D((float)(dd+x0+6*dx),(float)(y[0]+yi),(float)(dd+z0+6*dz));
			farleft[i] = new Point3D((float)(dd+x0-3*dx),(float)(y[0]+yi),(float)(dd+z0-3*dz));
			farright[i] = new Point3D((float)(dd+x0+8*dx),(float)(y[0]+yi),(float)(dd+z0+8*dz));
			if (i>0)
			{
				int k = (i-1)*8;
				int l = k+8;
				for (int j=0;j<7;j++)
				{
					setTriangle(nbt++,l+j,k+j,l+j+1);
					setTriangle(nbt++,l+j+1,k+j,k+j+1);
				}
			}
			a+=da;
			ay+=day;
			az+=daz;
		}
		int k = (div-1)*8;
		for (int j=0;j<7;j++)
		{
			setTriangle(nbt++,j,k+j,j+1);
			setTriangle(nbt++,j+1,k+j,k+j+1);
		}
		int tt[] = {6,0,2,2,2,4,6};
		for (int i=0;i<div;i++)
		{
			for (int j=0;j<7;j++)
				data[i][j] = (i<1 && j>1 && j<5)? 5 : tt[j]; 
		}
		computeTraj();
		initNormals();
	}

	public void computeTraj()
	{
		int j = 0;
		for (int i=0;i<traj.length;i++,j+=8)
		{
			traj[i] = new Point3D((vertices[j].x+vertices[j+7].x)/2,vertices[j].y+20,(vertices[j].z+vertices[j+7].z)/2);
		}
		for (int i=0;i<traj.length;i++)
		{
			j = i==0 ? traj.length-1 : i-1;
			roty[i] = (float) (Math.atan2(traj[i].x-traj[j].x,traj[i].z-traj[j].z));
		}
	}
	
	private static void saveCoords(PrintWriter pw,Point3D[] p) throws Exception
	{
		for (int i=0;i<p.length;i++)
			pw.println(p[i].x+" "+p[i].y+" "+p[i].z);
	}

	public void saveTrack(PrintWriter pw,boolean b) throws Exception
	{
		if (b)
		{
			for (int j=0;j<data.length;j++)
			{
				for (int k=0;k<data[j].length;k++)
					pw.print(data[j][k]+(k+1<data[j].length?" ":""));
				pw.println();
			}
		}
		saveCoords(pw,getVertices());
		saveCoords(pw,farleft);
		saveCoords(pw,left);
		saveCoords(pw,right);
		saveCoords(pw,farright);
	}

	private static void loadCoords(LineNumberReader lnr,Point3D[] p) throws Exception
	{
		for (int i=0;i<p.length;i++)
		{
			String s = lnr.readLine();
			int p0 = s.indexOf(' ');
			p[i].x = Float.valueOf(s.substring(0,p0)).floatValue();
			int p1 = s.indexOf(' ',p0+1);
			p[i].y = Float.valueOf(s.substring(p0+1,p1)).floatValue();
			p[i].z = Float.valueOf(s.substring(p1+1)).floatValue();
		}
	}

	public void loadTrack(LineNumberReader lnr,boolean b) throws Exception
	{
		if (b)
		{
			for (int j = 0; j < data.length; j++)
			{
				String sd = lnr.readLine();
				for (int k = 0; k < 7; k++)
				{
					int p = sd.indexOf(" ");
					data[j][k] = Integer.valueOf(sd.substring(0, p < 0 ? sd.length() : p)).intValue();
					sd = sd.substring(p + 1);
				}
			}
		}
		loadCoords(lnr,getVertices());
		loadCoords(lnr,farleft);
		loadCoords(lnr,left);
		loadCoords(lnr,right);
		loadCoords(lnr,farright);
		computeTraj();
		initNormals();
	}
}
