package j3d.basic.textured;

import j3d.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.util.*;

public class BasicTexturedObject3D {

  public static TexturedTriangle3D[] alloc(int tcount) {
    return new TexturedTriangle3D[tcount];
  }

  public static void setTriangle(Object3D obj,int idx,int a,int b,int c) {
    obj.getTriangles()[idx] = new TexturedTriangle3D(obj,a,b,c,(short[][][])null);
  }

  public static void setTexture(Triangle3D[] tab,Image texture) {
    setTexture(tab,new Image[] {texture});
  }

  public static void setTexture(Triangle3D[] tab,Image[] textures) {
    int w,h;
    BufferedImage[] tx = new BufferedImage[textures.length];
    Graphics2D[] g2d = new Graphics2D[textures.length];
    short[][][][] btx = new short[textures.length][][][];
    for (int i=0;i<textures.length;i++) {
        tx[i] = new BufferedImage(w=textures[i].getWidth(null),h=textures[i].getHeight(null),BufferedImage.TYPE_INT_ARGB);
        g2d[i] = (Graphics2D)tx[i].createGraphics();
        g2d[i].translate(w/2,h/2);
        g2d[i].rotate(Math.PI/2);
        g2d[i].drawImage(textures[i],-h/2,-w/2,null);
        btx[i] = TexturedTriangle3D.getTexture(tx[i]) ;
    }
    int k = 0;
    for (int i=0;i<tab.length;i++) {
        TexturedTriangle3D t = (TexturedTriangle3D)tab[i];
        t.setTexture(btx[k]);
        if (i%2==0) {
           t.tp[0] = new Point(0,0);
           t.tp[1] = new Point(0,tx[k].getHeight()-1);
           t.tp[2] = new Point(tx[k].getWidth()-1,0);
        } else {
           t.tp[0] = new Point(tx[k].getWidth()-1,tx[k].getHeight()-1);
           t.tp[1] = new Point(tx[k].getWidth()-1,0);
           t.tp[2] = new Point(0,tx[k].getHeight()-1);
           k = (k+1)%textures.length;
        }
    }
  }

  public static void setTexture(Object3D obj,Image texture) {
    Vector<Triangle3D> done = new Vector<Triangle3D>();
    Point3D[] v = obj.getVertices();
    Triangle3D[] t = obj.getTriangles();
    Point2D[][] points = new Point2D.Double[t.length][3];
    int[][] corr = new int[t.length][3];
    for (int i=0;i<t.length;i++) points[i]=new Point2D.Double[3];

    /* First segment */
    points[0][1] = new Point2D.Double(0.0,0.0);
    points[0][0] = new Point2D.Double(distance(v[t[0].a],v[t[0].b]),0.0);
    Vector<Object> stack = new Vector<Object>();
    pushData(stack,done,t,0,t[0].b,t[0].a,points[0][0],points[0][1]);

    /* launch mapping */
    while(stack.size()>0) {
       int nt = ((Integer)stack.elementAt(0)).intValue();
       int A = ((Integer)stack.elementAt(1)).intValue();
       int B = ((Integer)stack.elementAt(2)).intValue();
       Point2D pA = (Point2D)stack.elementAt(3);
       Point2D pB = (Point2D)stack.elementAt(4);
       for (int i=0;i<5;i++) stack.removeElementAt(0);
       plan(stack,done,nt,A,B,points,v,t,pA,pB,corr);
    }
    done.removeAllElements();

    /* apply to texture points */
    double minx = 1000,maxx=-1000;
    double miny = 1000,maxy=-1000;
    for (int i=0;i<points.length;i++) {
        for (int j=0;j<3;j++) {
            double x = points[i][j].getX();
            if (x<minx) minx=x;
            if (x>maxx) maxx=x;
            double y = points[i][j].getY();
            if (y<miny) miny=y;
            if (y>maxy) maxy=y;
        }
    }
    int w,h;
    BufferedImage tx = new BufferedImage(w=texture.getWidth(null),h=texture.getHeight(null),BufferedImage.TYPE_INT_ARGB);
    Graphics2D g2d = (Graphics2D)tx.createGraphics();
    g2d.drawImage(texture,0,0,null);
    double coefw = ((double)w-1)/(maxx-minx);
    double coefh = ((double)h-1)/(maxy-miny);
    short[][][] btext = TexturedTriangle3D.getTexture(tx);
    for (int i=0;i<t.length;i++) {
        TexturedTriangle3D tt = (TexturedTriangle3D)t[i];
        tt.setTexture(btext);
        for (int j=0;j<3;j++) tt.tp[j] = new Point((int)((points[i][corr[i][j]].getX()-minx)*coefw),(int)((points[i][corr[i][j]].getY()-miny)*coefh));
    }

    /* Show mapping in frame  */
    /*
    for (int i=0;i<t.length;i++) {
        for (int j=0;j<3;j++) points[i][j] = new Point2D.Double((points[i][j].getX()-minx)*coefw,(points[i][j].getY()-miny)*coefh);
    }
    xFrame f = new xFrame(points,texture);
    f.setSize(500,800);
    f.setVisible(true);
    */
  }

  protected static double distance(Point3D a,Point3D b) {
    return Math.sqrt((a.x-b.x)*(a.x-b.x)+(a.y-b.y)*(a.y-b.y)+(a.z-b.z)*(a.z-b.z));
  }

  protected static int findTriangle(Triangle3D[] t,int A,int B,Vector done) {
    for (int i=0;i<t.length;i++) {
        if (!done.contains(t[i])) {
           if ((t[i].a==A)&&((t[i].b==B)||(t[i].c==B))) return i;
           if ((t[i].b==A)&&((t[i].a==B)||(t[i].c==B))) return i;
           if ((t[i].c==A)&&((t[i].a==B)||(t[i].b==B))) return i;
        }
    }
    return -1;
  }

  protected static void pushData(Vector<Object> stack,Vector<Triangle3D> done,Triangle3D[] t,int nt,int A,int B,Point2D pA,Point2D pB) {
    done.add(t[nt]);
    stack.addElement(new Integer(nt));
    stack.addElement(new Integer(A));
    stack.addElement(new Integer(B));
    stack.addElement(pA);
    stack.addElement(pB);
  }

  protected static void plan(Vector<Object> stack,Vector<Triangle3D> done,int nt,int A,int B,Point2D[][] p,Point3D[] v,Triangle3D[] t,Point2D pA,Point2D pB,int[][] c) {
    Triangle3D tt = t[nt];
    int C = (tt.a==A ? (tt.b==B ? tt.c : tt.b) : (tt.b==A ? (tt.a==B ? tt.c : tt.a) : (tt.a==B ? tt.b : tt.a))),nnt;
    double AB = distance(v[A],v[B]);
    double BC = distance(v[B],v[C]);
    double AC = distance(v[A],v[C]);
    double tmp = Math.abs(pB.getY()-pA.getY())/AB;
    double thetaA = (tmp>=1?Math.PI/2:Math.asin(tmp));
    if(pA.getX()>pB.getX()) {
       if(pA.getY()<pB.getY()) thetaA=2*Math.PI-thetaA;
    } else {
       if(pA.getY()>pB.getY()) thetaA=Math.PI-thetaA;
       else thetaA+=Math.PI;
    }
    double aB = Math.acos((BC*BC+AB*AB-AC*AC)/(2*BC*AB));
    p[nt][0] = pA;
    p[nt][1] = pB;
    p[nt][2] = new Point2D.Double(pB.getX()+(Math.cos(thetaA-aB)*BC),pB.getY()+(Math.sin(thetaA-aB)*BC));
    c[nt][0] = (tt.a==A ? 0 : (tt.a==B ? 1 : 2));
    c[nt][1] = (tt.b==A ? 0 : (tt.b==B ? 1 : 2));
    c[nt][2] = (tt.c==A ? 0 : (tt.c==B ? 1 : 2));
    if((nnt = findTriangle(t,C,A,done))!=-1) pushData(stack,done,t,nnt,A,C,pA,p[nt][2]);
    if((nnt = findTriangle(t,B,C,done))!=-1) pushData(stack,done,t,nnt,C,B,p[nt][2],pB);
  }


  /* frame to show mapping result */
  /*
  private static class xFrame extends Frame {

    Point2D[][] points;
    Image tx;

    public xFrame(Point2D[][] p,Image img) { super(); points = p; tx = img;}

    public void paint(Graphics g) {
      g.drawImage(tx,50,50,null);
      g.setColor(Color.red);
      for (int i=0;i<points.length;i++) {
          g.drawLine(50+(int)points[i][0].getX(),50+(int)points[i][0].getY(),50+(int)points[i][1].getX(),50+(int)points[i][1].getY());
          g.drawLine(50+(int)points[i][1].getX(),50+(int)points[i][1].getY(),50+(int)points[i][2].getX(),50+(int)points[i][2].getY());
          g.drawLine(50+(int)points[i][2].getX(),50+(int)points[i][2].getY(),50+(int)points[i][0].getX(),50+(int)points[i][0].getY());
      }
    }
  }
  */

}
