package j3d.loader;

import java.io.*;
import java.util.*;
import java.awt.*;
import j3d.*;

public class ASCII3DSObject3DLoader implements Object3DLoader
{
	private String path;
	private String name;

	public ASCII3DSObject3DLoader()
	{
	}

	public Object3D loadObject(String fileName)
	{
		return loadObject(new File(fileName));
	}

	public Object3D loadObject(File file)
	{
		try
		{
			String s = name = file.getAbsolutePath();
			path = s.substring(0,s.lastIndexOf(File.separator)+1);
			return loadObject(new FileInputStream(file));
		}
		catch(Exception e) { e.printStackTrace(); }
    	return null;
	}

	public Object3D loadObject(InputStream is) throws IOException
	{
		Reader r = new BufferedReader(new InputStreamReader(is));
		StreamTokenizer st = new StreamTokenizer(r);
		st.eolIsSignificant(true);
		st.wordChars('{','{');
		st.wordChars('}','}');
		st.wordChars('_','_');
		st.wordChars('*','*');
		st.whitespaceChars(':',':');
		Vector<Vector<Object>> faces = new Vector<Vector<Object>>();
		Vector<Point3D> vertices = new Vector<Point3D>();
		Vector<Object> materials = new Vector<Object>();
		Vector<Point3D> tvertices = new Vector<Point3D>();
		String[] cur = new String[32];
		int lvl = 0;
		String lastWord = "";
		boolean bText = false;
		int deltaVert = 0;

		do
		{
			switch (st.nextToken())
			{
				case StreamTokenizer.TT_WORD:
					if (st.sval.equals("{"))
					{
						cur[lvl++] = lastWord;
						if (lastWord.equals("*MATERIAL"))
							bText = true;
					}
					else if (st.sval.equals("}"))
					{
						lvl-=1;
						if (cur[lvl].equals("*MATERIAL") && bText)
							materials.addElement(Color.blue);
						else if (cur[lvl].equals("*GEOMOBJECT"))
							tvertices.removeAllElements();

						if (lastWord.equals("*MESH_FACE"))
							deltaVert = vertices.size();
					}
					else if (st.sval.charAt(0)=='*')
						lastWord = st.sval;
					else if (lastWord.equals("*MESH_FACE"))
					{
						Vector<Object> face = new Vector<Object>();
						st.nextToken();
						face.addElement(new Integer((int)st.nval+deltaVert));
						st.nextToken();	// B
						st.nextToken();
						face.insertElementAt(new Integer((int)st.nval+deltaVert),0);
						st.nextToken();	// C
						st.nextToken();
						face.insertElementAt(new Integer((int)st.nval+deltaVert),0);
						faces.addElement(face);
						while (st.nextToken()!=StreamTokenizer.TT_EOL);
					}
					break;
				case StreamTokenizer.TT_NUMBER:
					if (lastWord.equals("*MESH_VERTEX"))
					{
						Point3D v = new Point3D();
						st.nextToken();
						v.z = (float)st.nval;
						st.nextToken();
						v.x = (float)st.nval;
						st.nextToken();
						v.y = (float)st.nval;
						vertices.addElement(v);
						while (st.nextToken()!=StreamTokenizer.TT_EOL);
					}
					else if (lastWord.equals("*MESH_TVERT"))
					{
// TO DO : multiple objects textures
						Point3D v = new Point3D();
						st.nextToken();
						v.x = (float)st.nval;
						st.nextToken();
						v.y = (float)st.nval;
						st.nextToken();
						v.z = (float)st.nval;
						tvertices.addElement(v);
						while (st.nextToken()!=StreamTokenizer.TT_EOL);
					}
					else if (lastWord.equals("*MESH_TFACE"))
					{
// TO DO : multiple objects textures
						Vector<Object> face = faces.elementAt((int)st.nval);
						Vector<Integer> tface = new Vector<Integer>();
						face.addElement(tface);
						st.nextToken();
						tface.addElement(new Integer((int)(st.nval)));
						st.nextToken();
						tface.addElement(new Integer((int)(st.nval)));
						st.nextToken();
						tface.addElement(new Integer((int)(st.nval)));
						while (st.nextToken()!=StreamTokenizer.TT_EOL);
					}
					else if (lastWord.equals("*MATERIAL_REF"))
					{
						Object mat = materials.elementAt((int)st.nval);
						if (mat instanceof Color)
						{
							for (int i=0;i<faces.size();i++)
								faces.elementAt(i).addElement((Color)mat);
						}
						else
						{
							short[][][] text = (short[][][])mat;
							int th = text.length-1;
							int tw = text[0].length-1;
							for (int i=0;i<faces.size();i++)
							{
								Vector<Object> face = faces.elementAt(i);
								Vector tface = (Vector)face.elementAt(3);
								Vector<Object> tdata = new Vector<Object>();
								for (int j=0;j<3;j++)
								{
									int v = ((Integer)tface.elementAt(j)).intValue();
									Point3D p = (Point3D)tvertices.elementAt(v);
									tdata.insertElementAt(new Integer((int)(p.x*tw)),0);
									tdata.insertElementAt(new Integer((int)(p.y*th)),1);
								}
								tdata.insertElementAt(text,0);
								face.removeElementAt(3);
								face.addElement(tdata);
							}
						}
					}
					break;
				case '"':
					if (lastWord.equals("*BITMAP"))
					{
						try
						{
							materials.addElement(TexturedTriangle3D.getTexture(path+st.sval));
							bText = false;
						}
						catch(Exception e)
						{
							e.printStackTrace();
						}
					}
				case StreamTokenizer.TT_EOL:
				case StreamTokenizer.TT_EOF:
				default:
					break;
			}
		} while (st.ttype != StreamTokenizer.TT_EOF);
		is.close();
		return new DefaultObject3D(name,vertices,faces);
	}

	public Object3D loadObject(Object source)
	{
		return null;
	}
}