package j3d.loader;

import j3d.DefaultScene3D;
import j3d.Object3D;
import j3d.Point3D;
import j3d.Position3D;
import j3d.Rotation3D;
import j3d.TexturedTriangle3D;

import java.awt.Canvas;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.util.Hashtable;
import java.util.Vector;

public class DefaultObject3DLoader implements Object3DLoader
{
	private String name = "unknown";
	private Hashtable<String,short[][][]> materials;
	
	public DefaultObject3DLoader()
	{
	}
	
	public Object3D loadObject(String fileName)
	{
		return loadObject(new File(fileName));
	}
	
	public Object3D loadObject(File file)
	{
		try
		{
			name = file.getAbsolutePath();
			return loadObject(new FileInputStream(file));
		} catch(Exception e) { e.printStackTrace(); }
		return null;
	}

	private static Hashtable<String,Image> m_hImg = new Hashtable<String,Image>();

	public static BufferedImage loadImage(String n)
	{
		BufferedImage img = (BufferedImage)m_hImg.get(n);
		if (img!=null)
			return img;
		System.err.println("loading "+n);
		try
		{
			if (n.endsWith(".PCX") || n.endsWith(".pcx"))
				img = (BufferedImage)PCXDecoder.getImage(n);
			else if (n.endsWith(".TGA") || n.endsWith(".tga"))
				img = (BufferedImage)TGADecoder.getImage(n);
			else if (n.endsWith(".GIF") || n.endsWith(".gif") ||
					n.endsWith(".JPG") || n.endsWith(".jpg") ||
					n.endsWith(".PNG") || n.endsWith(".png"))
			{
				Image image = Toolkit.getDefaultToolkit().createImage(n);
				MediaTracker mt = new MediaTracker(new Canvas());
				mt.addImage(image,0);
				mt.waitForAll();
				img = new BufferedImage(image.getWidth(null),image.getHeight(null),BufferedImage.TYPE_INT_ARGB);
				img.getGraphics().drawImage(image,0,0,null);
				m_hImg.put(n,img);
			}
			System.err.println(img.getWidth()+"x"+img.getHeight());
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		return img;
	}
	
	private void loadMaterials(String mat) throws IOException
	{
		File parent = (new File(name)).getParentFile();
		File f = new File(parent,mat);
		Reader r = new BufferedReader(new InputStreamReader(new FileInputStream(f)));
		StreamTokenizer st = new StreamTokenizer(r);
		st.eolIsSignificant(true);
//		st.commentChar('#');
		st.wordChars('/','/');
		st.wordChars('#','#');
		st.wordChars('_','_');
		boolean cont = true;
		String matname = "";
		while (cont)
		{
			switch (st.nextToken())
			{
				default:
					break;
				case StreamTokenizer.TT_EOF:
					cont = false;
				break;
				case StreamTokenizer.TT_EOL:
					break;
				case StreamTokenizer.TT_WORD:
					if ("newmtl".equals(st.sval))
					{
						matname="";
						st.nextToken();
						while (st.ttype!=StreamTokenizer.TT_EOL)
						{
							matname+=(matname.length()==0?"":" ")+(st.ttype==StreamTokenizer.TT_WORD?st.sval:String.valueOf((int)st.nval));
							st.nextToken();
							System.err.println(matname);
						}
					}
					else if ("map_Kd".equals(st.sval))
					{
						st.nextToken();
						File pix = new File(parent,st.sval);
						String path = pix.getAbsolutePath();
						materials.put(matname,TexturedTriangle3D.getTexture(path,loadImage(path)));
					}
				break;
			}
		}
	}

	public Object3D loadObject(InputStream is) throws IOException
	{
		materials = new Hashtable<String,short[][][]>();
		Reader r = new BufferedReader(new InputStreamReader(is));
		StreamTokenizer st = new StreamTokenizer(r);
		st.eolIsSignificant(true);
		st.commentChar('#');
		st.ordinaryChar('/');
		st.wordChars('_','_');
		DefaultScene3D scene = new DefaultScene3D(name);
		Vector<Vector<Object>> faces = new Vector<Vector<Object>>();
		Vector<Point3D> vertices = new Vector<Point3D>();
		Vector<Point3D> vt = new Vector<Point3D>();
		Vector<Vector<Object>> old_faces=null;
		Vector<Point3D> old_vertices=null;
		Vector<Point3D> old_vt=null;
		Integer curColor = null;
		int ccol = 0,cnt=0;
		boolean negmode = false;
		short[][][] curmat = null;
		scan:
			while (faces.size()<300000)
			{
				switch (st.nextToken())
				{
				default:
					break scan;
				case StreamTokenizer.TT_EOL:
					break;
				case StreamTokenizer.TT_WORD:
					if ("v".equals(st.sval))
					{
						double x = 0, y = 0, z = 0;
						if (st.nextToken() == StreamTokenizer.TT_NUMBER)
						{
							x = st.nval;
							if (st.nextToken() == StreamTokenizer.TT_NUMBER)
							{
								y = st.nval;
								if (st.nextToken() == StreamTokenizer.TT_NUMBER)
									z = st.nval;
							}
						}
						vertices.add(new Point3D((float)x,(float)y,(float)z));
						while (st.nextToken() != StreamTokenizer.TT_EOL && st.ttype != StreamTokenizer.TT_EOF);
					}
					else if ("vt".equals(st.sval))
					{
						Point3D p = new Point3D();
						st.nextToken();
						p.x = (float)st.nval;
						st.nextToken();
						p.y = (float)st.nval;
						if (p.y<0)
						{
							p.y = 1+p.y;
							//p.x = 1-p.x;
						}
						vt.addElement(p);
						while (st.nextToken() != StreamTokenizer.TT_EOL && st.ttype != StreamTokenizer.TT_EOF);
					}
					else if ("f".equals(st.sval) || "fo".equals(st.sval) || "l".equals(st.sval))
					{
						Vector<Object> face = new Vector<Object>();
						Vector<Object> att = null;
						while (true)
						{
							if (st.nextToken() == StreamTokenizer.TT_NUMBER)
							{
								int nv = (int)st.nval;
								if (nv<0)
								{
									negmode = true;
									if (vertices.size()==0)
									{
										vt = old_vt;
										vertices = old_vertices;
										faces = old_faces;
									}
									nv = vertices.size()+nv;
								} else nv -= 1;
								face.insertElementAt(new Integer(nv),0);
							}
							else if (st.ttype == '/')
							{
								st.nextToken(); // vertex texture
								if (curmat!=null)
								{
									if (att==null)
									{
										att = new Vector<Object>();
										att.addElement(curmat);
									}
									Point3D p = (Point3D)vt.elementAt((int)st.nval-1);
									int x = (int)(p.x*curmat[0].length);
									int y = (int)((p.y+1)*curmat.length/2);
									if (x<0)
										x=0;
									else if (x>=curmat[0].length)
										x=curmat[0].length-1;
									if (y<0)
										y=0;
									else if (y>=curmat.length)
										y=curmat.length-1;
									att.addElement(new Integer(x));
									att.addElement(new Integer(y));
									st.nextToken(); // another /
									st.nextToken(); // vertex normal
								}
							}
							else break;
						}
						if (face.size()>3)
						{
							Vector<Object> f3 = new Vector<Object>();
							f3.addElement(face.elementAt(0));
							f3.addElement(face.elementAt(1));
							f3.addElement(face.elementAt(2));
							f3.addElement(curColor);
							faces.addElement(f3);
							if (face.size()==4)
							{
								f3 = new Vector<Object>();
								f3.addElement(face.elementAt(0));
								f3.addElement(face.elementAt(2));
								f3.addElement(face.elementAt(3));
								f3.addElement(curColor);
								faces.addElement(f3);
							}
							if (face.size()==5)
							{
								f3 = new Vector<Object>();
								f3.addElement(face.elementAt(0));
								f3.addElement(face.elementAt(3));
								f3.addElement(face.elementAt(4));
								f3.addElement(curColor);
								faces.addElement(f3);
							}
							if (face.size()==6)
							{
								f3 = new Vector<Object>();
								f3.addElement(face.elementAt(0));
								f3.addElement(face.elementAt(4));
								f3.addElement(face.elementAt(5));
								f3.addElement(curColor);
								faces.addElement(f3);
							}
						}
						else
						{
							face.addElement(att==null?(Object)curColor:(Object)att);
							faces.addElement(face);
						}
						if (st.ttype != StreamTokenizer.TT_EOL)
							break scan;
					}
					else if ("g".equals(st.sval) || "o".equals(st.sval))
					{
						if (negmode)
						{
							if (faces.size()>0)
								scene.addObject(new DefaultObject3D(name+" "+(cnt++),vertices,faces),new Position3D(new Point3D(0,0,0),new Rotation3D(0,0,0),1));
							old_vt = vt;
							old_vertices = vertices;
							old_faces = faces;
							vt = new Vector<Point3D>();
							vertices = new Vector<Point3D>();
							faces = new Vector<Vector<Object>>();
						}
						st.nextToken();
						if (st.ttype == StreamTokenizer.TT_NUMBER)
						{
							curColor=new Integer((int)st.nval);
						}
						else if (st.ttype == StreamTokenizer.TT_WORD && materials.get(st.sval)!=null)
						{
							curmat = (short[][][])materials.get(st.sval);
							if (curmat==null)
								curColor=new Integer(ccol++);
						}
						else
							curColor=new Integer(ccol++);
						while (st.nextToken() != StreamTokenizer.TT_EOL && st.ttype != StreamTokenizer.TT_EOF) ;
					}
					else if ("mtllib".equals(st.sval))
					{
						st.nextToken();
						loadMaterials(st.sval);
					}
					else if ("usemtl".equals(st.sval))
					{
						st.nextToken();
						curmat = (short[][][])materials.get(st.sval);
					}
					else
					{
						while (st.nextToken() != StreamTokenizer.TT_EOL && st.ttype != StreamTokenizer.TT_EOF);
					}
				}
			}
		is.close();
		if (faces.size()>0)
			scene.addObject(new DefaultObject3D(name+" "+(cnt++),vertices,faces),new Position3D(new Point3D(0,0,0),new Rotation3D(0,0,0),1));
		//if (st.ttype != StreamTokenizer.TT_EOF) throw new IOException(st.toString());
		return scene;
	}
	
	public Object3D loadObject(Object source)
	{
		return null;
	}
}
