package j3d.loader;

import j3d.Object3D;
import j3d.Point3D;
import j3d.TexturedTriangle3D;

import java.awt.Color;
import java.awt.Point;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Hashtable;
import java.util.Vector;

public class LWOObject3DLoader
{
	private String name;
	private String path;
	private byte[] binlwo;
	private int offset;
	private Vector<String> objs = null;
	private Vector<Point3D> pnts = null;
	private Vector<String> srfs = null;
	private Vector<Vector<Object>> pols = null;
	private Hashtable<String,Hashtable<String,Object>> surf = null;

	public LWOObject3DLoader()
	{
	}

	public Object3D loadObject(String fileName)
	{
		return loadObject(new File(fileName));
	}

	public Object3D loadObject(File file)
	{
		try
		{
			name = file.getAbsolutePath();
			path = file.getParentFile().getAbsolutePath()+File.separatorChar;
			binlwo = new byte[(int)file.length()];
			return loadObject(new FileInputStream(file));
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		return null;
	}

	/*
	private void search(String chunk)
	{
		String s = "";
		s+=(char)binlwo[offset++];
		s+=(char)binlwo[offset++];
		s+=(char)binlwo[offset++];
		s+=(char)binlwo[offset++];
		while (!s.equals(chunk))
			s=s.substring(1)+(char)binlwo[offset++];
	}
	*/

	private int readChar()
	{
		if (offset==binlwo.length)
			return -1;
		int b = binlwo[offset++];
		return(b<0?256:0)+b;
	}

	private int readInt()
	{
		int i = readChar();
		i = (i<<8)|readChar();
		return i;
	}

	private int readLong()
	{
		int i = readChar();
		i = (i<<8)|readChar();
		i = (i<<8)|readChar();
		i = (i<<8)|readChar();
		return i;
	}

	private float readFloat()
	{
		int i = readLong();
		return Float.intBitsToFloat(i);
	}
	
	private String readName()
	{
		String s = "";
		int c;
		while((c = readChar())>0)
			s += (char)c;
		while(binlwo[offset]==0)
			offset+=1;
		return s;
	}

	private Hashtable<String,Object> readSurfChunk(int end)
	{
		Hashtable<String,Object> h = new Hashtable<String,Object>();
		while (offset<end)
		{
			String cname = "";
			cname+=(char)readChar();
			cname+=(char)readChar();
			cname+=(char)readChar();
			cname+=(char)readChar();
			int csize = readInt();
			if (cname.equals("COLR")||cname.equals("TCLR"))
			{
				h.put(cname,new Color(readChar(),readChar(),readChar()));
				readChar();
			}
			else if (cname.equals("FLAG")||cname.equals("LUMI")||cname.equals("DIFF")||
					 cname.equals("SPEC")||cname.equals("REFL")||cname.equals("TRAN")||
					 cname.equals("GLOS")||cname.equals("TFLG")||cname.equals("TVAL")||
					 cname.equals("TFRQ"))
			{
				h.put(cname,new Integer(readInt()));
			}
			else if (cname.equals("RIMG")||cname.equals("CTEX")||cname.equals("DTEX")||
					 cname.equals("STEX")||cname.equals("RTEX")||cname.equals("TTEX")||
					 cname.equals("BTEX")||cname.equals("TIMG")||cname.equals("TALP"))
			{
				String s = readName();
				if (cname.equals("TIMG"))
				{
					int pos = s.lastIndexOf('\\');
					if (pos<0)
						pos = s.lastIndexOf('/');
					BufferedImage img = DefaultObject3DLoader.loadImage(path+(pos<0?s:s.substring(pos+1)));
					if (img!=null)
						h.put(cname,TexturedTriangle3D.getTexture(s,img));
				}
				else
					h.put(cname,s);
			}
			else if (cname.equals("RSAN")||cname.equals("RIND")||cname.equals("EDGE")||
					 cname.equals("SMAN")||cname.equals("TAMP")||cname.equals("TAAS")||
					 cname.equals("VLUM")||cname.equals("VDIF")||cname.equals("VSPC")||
					 cname.equals("VRFL")||cname.equals("VTRN")||cname.equals("TOPC")||
					 cname.equals("TSP0")||cname.equals("TSP1")||cname.equals("TSP2"))
			{
				h.put(cname,new Float(readFloat()));
			}
			else if (cname.equals("TSIZ")||cname.equals("TCTR")||
					 cname.equals("TFAL")||cname.equals("TVEL"))
			{
				h.put(cname,new Point3D(readFloat(),readFloat(),readFloat()));
			}
			else if (cname.equals("TWRP"))
			{
				h.put(cname,new Point(readInt(),readInt()));
			}
			else
			{
				System.err.println("Skipping subchunk "+cname+" "+csize);
				offset+=csize;
			}
		}
System.err.println(h.keySet());
		return h;
	}

	private void readChunk(int end)
	{
		int dp = 0;
		int ds = 0;
		while (offset<end)
		{
			String cname = "";
			cname+=(char)readChar();
			cname+=(char)readChar();
			cname+=(char)readChar();
			cname+=(char)readChar();
			int csize = readLong();
			int cend = offset+csize;
System.err.println(cname+" "+Integer.toHexString(offset)+" "+Integer.toHexString(cend));
			if (cname.equals("FORM"))
			{
				offset+=4; // LWOB
				readChunk(cend);
			}
			else if (cname.equals("TAGS"))
			{
				objs = new Vector<String>();
				while(offset<cend)
					objs.addElement(readName());
			}
			else if (cname.equals("LAYR"))
			{
				offset+=csize;
				dp = pnts==null?0:pnts.size();
				ds = srfs==null?0:srfs.size();
			}
			else if (cname.equals("PNTS"))
			{
				if (pnts==null)
					pnts = new Vector<Point3D>();
				while(offset<cend)
					pnts.addElement(new Point3D(readFloat(),readFloat(),readFloat()));
			}
			else if (cname.equals("SRFS"))
			{
				if (srfs==null)
					srfs = new Vector<String>();
				while(offset<cend)
					srfs.addElement(readName());
			}
			else if (cname.equals("POLS"))
			{
				boolean b = false;
				if (binlwo[offset]  =='F' && binlwo[offset+1]=='A' &&
					binlwo[offset+1]=='C' && binlwo[offset+1]=='E')
				{
					offset+=4;
					b = true;
				}
				if (pols==null)
					pols = new Vector<Vector<Object>>();
				while(offset<cend)
				{
					Vector<Object> p = new Vector<Object>();
					int n = readInt();
					for (int i=0;i<n;i++)
						p.insertElementAt(new Integer(dp+readInt()),0);
					Integer srf = new Integer(ds+(b?0:readInt()));
					if (p.size()<3)
						continue;
					else if (p.size()>3)
					{
//System.err.println(p.size());
						for (int i=1;i<p.size()-1;i++)
						{
							Vector<Object> f3 = new Vector<Object>();
							f3.addElement(p.elementAt(0));
							f3.addElement(p.elementAt(i));
							f3.addElement(p.elementAt(i+1));
							f3.addElement(srf);
							pols.addElement(f3);
						}
					}
					else
					{
						p.addElement(srf);
						pols.addElement(p);
					}
				}
			}
			else if (cname.equals("SURF"))
			{
				String n = readName();
				if (surf==null)
					surf = new Hashtable<String,Hashtable<String,Object>>();
				surf.put(n,readSurfChunk(cend));
			}
			else
			{
				System.err.println("Skipping chunk "+cname+" "+csize);
				offset+=csize;
			}
		}
	}

	public Object3D loadObject(InputStream stream) throws IOException
	{
		stream.read(binlwo);
		offset = 0;
		readChunk(binlwo.length);
		for (int i=0;i<pols.size();i++)
		{
			Vector<Object> f = pols.elementAt(i);
			int j = f.size()-1;
			Hashtable<String,Object> h = surf==null?null:surf.get(srfs.elementAt(((Integer)f.elementAt(j)).intValue()-1));
			f.removeElementAt(j);
			short[][][] tex = h==null?null:(short[][][])h.get("TIMG");
			if (tex!=null)
			{
				Vector<Object> att = new Vector<Object>(7);
				att.addElement(tex);
				att.addElement(new Integer(0));
				att.addElement(new Integer(0));
				att.addElement(new Integer(tex[0].length-1));
				att.addElement(new Integer(0));
				att.addElement(new Integer(0));
				att.addElement(new Integer(tex.length-1));
				//f.addElement(att);
				f.addElement(h==null?Color.blue:h.get("COLR"));
			}
			else
				f.addElement(h==null?Color.blue:h.get("COLR"));
		}
		return new DefaultObject3D(name,pnts,pols);
	}
}
