package j3d.loader;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.image.*;

import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import j3d.*;

public class MD2Object3DLoader implements Object3DLoader,Object3D
{
	private String path;
	private String name,_name;
	private DefaultObject3D[] frames;
	private int curframe;
	ArrayList<String> names;
	int[] anims;
	boolean[] export;
	private BufferedImage m_skin0;
	private float m_w0,m_h0;

	public MD2Object3DLoader()
	{
	}

	public Object3D loadObject(String fileName)
	{
		return loadObject(new File(fileName));
	}

	private static int readInt(InputStream is) throws IOException
	{
		int res = is.read() + (is.read()<<8) + (is.read()<<16) + (is.read()<<24);
    	return res;
	}

	private static int readShort(InputStream is) throws IOException
	{
		int res = is.read() + (is.read()<<8);
		return res;
	}

	private static float readFloat(InputStream is) throws IOException
	{
		return Float.intBitsToFloat(readInt(is));
	}

	private static int btoi(byte b)
	{
		int a = b;
		return (a<0?256+a:a);
	}

	public Object3D loadObject(File file)
	{
		try
		{
			String s = file.getAbsolutePath();
			name = s.substring(s.lastIndexOf(File.separator)+1,s.length()-4);
			path = s.substring(0,s.lastIndexOf(File.separator));
			if (name.toLowerCase().equals("tris"))
			{
				name = path.substring(path.lastIndexOf(File.separator)+1);
				_name = "tris";
			}
			else
				_name = name;
			path+=File.separator;
			return loadObject(new FileInputStream(file));
		}
		catch(Exception e) { e.printStackTrace(); }
    	return null;
	}

	private void tabRemove(Object[] tab,int n,boolean[] u)
	{
		for (int i=n;i<tab.length-1;i++)
		{
			tab[i] = tab[i+1];
			if (u!=null)
				u[i] = u[i+1];
		}
	}

	private BufferedImage findImage(String n)
	{
System.err.println(n);
		if (!(new File(n)).exists())
			return null;
		BufferedImage img = DefaultObject3DLoader.loadImage(n);
System.err.println(img.getWidth()+" "+img.getHeight());
		if (m_skin0==null)
		{
			m_skin0 = new BufferedImage(128,128,BufferedImage.TYPE_INT_ARGB);
			m_skin0.getGraphics().drawImage(img,0,0,128,128,0,0,img.getWidth(),img.getHeight(),null);
			m_w0 = 128f/img.getWidth();
			m_h0 = 128f/img.getHeight();
		}
		return img;
	}

	public void center()
	{
		for (int i=0;i<frames.length;i++)
			frames[i].center();
	}

	public Object3D loadObject(InputStream is) throws IOException
	{
		ArrayList<ArrayList<Object>> faces = new ArrayList<ArrayList<Object>>();
		ArrayList<Point3D> vertices = new ArrayList<Point3D>();

		readInt(is); // ID = "IPD2"
		readInt(is); // version = 8
		/*int skinwidth = */readInt(is);
		/*int skinheight = */readInt(is);
    	/*int framesize = */readInt(is);
		int num_skins = readInt(is);
		int num_xyz = readInt(is);
		int num_st = readInt(is);
		int num_tris = readInt(is);
		/*int num_glcmds = */readInt(is);
		int num_frames = readInt(is);
		/*int ofs_skins = */readInt(is);
		/*int ofs_st = */readInt(is);
		/*int ofs_tris = */readInt(is);
		/*int ofs_frames = */readInt(is);
		/*int ofs_glcmds = */readInt(is);
		/*int ofs_end = */readInt(is);

		int i,j;

		// skin names
		short[][][][] skins = new short[num_skins==0?1:num_skins][][][];
		for (i=0;i<num_skins;i++)
		{
			byte[] b = new byte[64];
			is.read(b);
			String n = new String(b);
			n = n.substring(0,n.indexOf(0));
			if (n.indexOf('/')>=0)
				n = n.substring(n.lastIndexOf('/')+1);
			else
				n = n.substring(n.lastIndexOf('\\')+1);
			try
			{
				BufferedImage img = findImage(path+n);
				if (img==null && n.startsWith("w_"))
					img = findImage(path+"weapon.pcx");
				if (img != null)
					skins[i] = TexturedTriangle3D.getTexture(path+n,img);
			}
			catch(Exception e)
			{
				System.err.println("Can not load "+path+n+" :"+e.getMessage());
			}
		}
		if (num_skins==0 || skins[0]==null)
		{
			String[] exts = new String[] {".pcx",".tga",".gif",".png",".jpg"};
			BufferedImage img = null;
			for (i=0;i<exts.length && img==null;i++)
				img = findImage(path+name+exts[i]);
			if (img==null && name.startsWith("w_"))
				img = findImage(path+"weapon.pcx");
			skins[0] = TexturedTriangle3D.getTexture(path+name+exts[i-1],img);
		}

		// texture coords
		int[][] tab_st = new int[num_st][2]; // num_st==num_xyz ?
		for (i=0;i<num_st;i++)
		{
			tab_st[i][0] = readShort(is);
			tab_st[i][1] = readShort(is);
		}

		// triangles
		int[][][] tab_tris = new int[num_tris][2][3];
		for (i=0;i<num_tris;i++)
		{
			// vertices
			tab_tris[i][0][0] = readShort(is);
			tab_tris[i][0][1] = readShort(is);
			tab_tris[i][0][2] = readShort(is);
			// texture index
			tab_tris[i][1][0] = readShort(is);
			tab_tris[i][1][1] = readShort(is);
			tab_tris[i][1][2] = readShort(is);
		}

		// frames

		float[][] tab_frames_scale = new float[num_frames][3];
		float[][] tab_frames_translate = new float[num_frames][3];
		byte[][] tab_frames_name = new byte[num_frames][16];
		byte[][][] tab_frames_verts = new byte[num_frames][num_xyz][4];
		Point3D[][] pf = new Point3D[num_frames][num_xyz];

		int anim = -1;
		String prefix = "[SoD]";
		names = new ArrayList<String>();
		anims = new int[num_frames];
		export = new boolean[num_frames];

		for (i=0;i<num_frames;i++)
		{
			tab_frames_scale[i][0] = readFloat(is);
			tab_frames_scale[i][1] = readFloat(is);
			tab_frames_scale[i][2] = readFloat(is);
			tab_frames_translate[i][0] = readFloat(is);
			tab_frames_translate[i][1] = readFloat(is);
			tab_frames_translate[i][2] = readFloat(is);

			// animation name
			is.read(tab_frames_name[i]);
			String fname = new String(tab_frames_name[i]);
			fname = fname.substring(0,fname.indexOf(0));
			if (fname.startsWith(prefix))
				anims[i] = anim;
			else
			{
				int l = fname.length();
				int pos = l-((fname.charAt(l-1)=='0'||fname.charAt(l-1)=='1')?1:0);
				pos -= (fname.charAt(l-2)=='0')?1:0;
				prefix = fname.substring(0,pos);
System.err.println(prefix);
				anims[i] = ++anim;
				names.add(prefix);
			}
			export[i] = true;

			for (j=0;j<num_xyz;j++)
			{
				is.read(tab_frames_verts[i][j]);
				pf[i][j] = new Point3D(	btoi(tab_frames_verts[i][j][1])*tab_frames_scale[i][1]+tab_frames_translate[i][1],
										btoi(tab_frames_verts[i][j][2])*tab_frames_scale[i][2]+tab_frames_translate[i][2],
										btoi(tab_frames_verts[i][j][0])*tab_frames_scale[i][0]+tab_frames_translate[i][0]);

			}
		}

		// triangle count reduction

		float[] l = new float[num_tris];
		Point3D pt = new Point3D();
		int min=-1;
		int max_tris = (num_tris*100)/100;
		boolean[] used = new boolean[num_tris];
		for (i=0;i<num_tris;i++)
			used[i] = false;
System.err.println("f="+num_frames+",t="+num_tris+",v="+num_xyz);
		while (num_tris>max_tris)
		{
			min = -1;
			for (i=0;i<num_tris;i++)
			{
				if (used[i])
					continue;
				Point3D p0 = pf[0][tab_tris[i][0][0]];
				Point3D p1 = pf[0][tab_tris[i][0][1]];
				Point3D p2 = pf[0][tab_tris[i][0][2]];
				pt.x = (p0.x+p1.x+p2.x)/3;
				pt.y = (p0.y+p1.y+p2.y)/3;
				pt.z = (p0.z+p1.z+p2.z)/3;
				l[i] = pt.dist(p0)+pt.dist(p1)+pt.dist(p2);
				if (min<0 || l[i]<l[min])
					min = i;
			}
			if (min==-1)
			{
				for (i=0;i<num_tris;i++)
					used[i] = false;
				continue;
			}
			int a = tab_tris[min][0][0];
			int b = tab_tris[min][0][1];
			int c = tab_tris[min][0][2];
			int d;
			if (b<a) { d = a; a = b; b = d; }
			if (c<a) { d = a; a = c; c = d; }
			if (c<b) { d = b; b = c; c = d; }
			num_xyz-=2;
			for (i=0;i<num_frames;i++)
			{
				Point3D np = new Point3D(	(pf[i][a].x+pf[i][b].x+pf[i][c].x)/3,
											(pf[i][a].y+pf[i][b].y+pf[i][c].y)/3,
											(pf[i][a].z+pf[i][b].z+pf[i][c].z)/3);
				tabRemove(pf[i],c,null);
				tabRemove(pf[i],b,null);
				tabRemove(pf[i],a,null);
				pf[i][num_xyz-1] = np;
			}
			for (i=0;i<num_tris;i++)
			{
				for (j=0;j<3;j++)
				{
					if (tab_tris[i][0][j]==a || tab_tris[i][0][j]==b || tab_tris[i][0][j]==c)
					{
						tab_tris[i][0][j] = num_xyz-1;
						used[i] = true;
					}
					else if (tab_tris[i][0][j]>c)
						tab_tris[i][0][j]-=3;
					else if (tab_tris[i][0][j]>b)
						tab_tris[i][0][j]-=2;
					else if (tab_tris[i][0][j]>a)
						tab_tris[i][0][j]-=1;
				}
			}
			for (i=num_tris-1;i>=0;i--)
			{
				if (tab_tris[i][0][0]==tab_tris[i][0][1] || tab_tris[i][0][0]==tab_tris[i][0][2] ||	tab_tris[i][0][1]==tab_tris[i][0][2])
				{
					tabRemove(tab_tris,i,used);
					num_tris-=1;
				}
			}
		}

System.err.println("->"+num_tris+","+num_xyz);

		// build objects
		frames = new DefaultObject3D[num_frames];
		for (i=0;i<num_frames;i++)
		{
			faces.clear();
			vertices.clear();
			for (j=0;j<num_xyz;j++)
				vertices.add(pf[i][j]);

			for (j=0;j<num_tris;j++)
			{
				ArrayList<Object> v = new ArrayList<Object>();
				v.add(new Integer(tab_tris[j][0][0]));
				v.add(new Integer(tab_tris[j][0][1]));
				v.add(new Integer(tab_tris[j][0][2]));
				ArrayList<Object> att = new ArrayList<Object>();
				att.add(skins[0]);
				att.add(new Integer(tab_st[tab_tris[j][1][0]][0]));
				att.add(new Integer(tab_st[tab_tris[j][1][0]][1]));
				att.add(new Integer(tab_st[tab_tris[j][1][1]][0]));
				att.add(new Integer(tab_st[tab_tris[j][1][1]][1]));
				att.add(new Integer(tab_st[tab_tris[j][1][2]][0]));
				att.add(new Integer(tab_st[tab_tris[j][1][2]][1]));
				v.add(att);
				faces.add(v);
			}
			frames[i] = new DefaultObject3D(name+"_"+i,vertices,faces);
		}
		curframe = 0;
		return this;
	}

	public Object3D loadObject(Object source)
	{
		return null;
	}

	/* Object3D interface */

	public Point3D[] getVertices()
	{
		return frames[curframe].getVertices();
	}

	public Triangle3D[] getTriangles()
	{
		return frames[curframe].getTriangles();
	}

	public Vector3D[] getVNormals()
	{
		return frames[curframe].getVNormals();
	}

	public Vector3D[] getTNormals()
	{
		return frames[curframe].getTNormals();
	}

	public Point3D[] getBoundingBox()
	{
		return frames[curframe].getBoundingBox();
	}

	public Object3D copy()
	{
		return frames[curframe].copy();
	}

	public void transform(Object3D o,Matrix3D cur,Position3D pos,Point3D[] v,Triangle3D[] t,Vector3D[] vnorm,Vector3D[] tnorm,Point dVT)
	{
		frames[curframe].transform(o,cur,pos,v,t,vnorm,tnorm,dVT);
	}

	public String toString()
	{
		return path+_name+".md2";
	}

	/* Animation */

	public ArrayList<String> getNames()
	{
		return names;
	}

	public void enableExport(int n)
	{
		for (int i=0;i<frameCount();i++)
			if (n==-1 || anims[i]==n)
				export[i] = true;
	}

	public int exportIndex(String name)
	{
		int n = names.indexOf(name),i,j=0;
		for (i=0;anims[i]!=n;i++)
		{
			if (export[i])
				j++;
		}
		return j;
	}

	public int frameCount(String name)
	{
		int n = names.indexOf(name),i,j=0;
		for (i=0;i<frameCount();i++)
		{
			if (anims[i]==n)
				j++;
		}
		return j;
	}

	public int exportIndexEnd(String name)
	{
		int n =  names.indexOf(name),i,j=0;
		for (i=0;anims[i]!=n;i++)
		{
			if (export[i])
				j++;
		}
		for (;i<anims.length && anims[i]==n;i++)
		{
			if (export[i])
				j++;
		}
		return j-1;
	}

	public void disableExport(int n)
	{
		for (int i=0;i<frameCount();i++)
			if (n==-1 || anims[i]==n)
				export[i] = false;
	}

	public void nextFrame()
	{
		do
		{
			if (++curframe==frameCount())
				curframe=0;
		} while (!export[curframe]);
	}

	public void previousFrame()
	{
		do
		{
			if (--curframe<0)
				curframe=frameCount()-1;
		} while (!export[curframe]);
	}

	public int frameCount()
	{
		return frames.length;
	}

	public int exportFrameCount()
	{
		int n = 0;
		for (int i=0;i<frameCount();i++)
			if (export[i])
				n+=1;
		return n;
	}

	public void setFrame(int num)
	{
		curframe = num-1;
		nextFrame();
	}

	public DefaultObject3D getFrame(int num)
	{
		return frames[num];
	}

	public DefaultObject3D getFrame()
	{
		return frames[curframe];
	}
	
	private static void write16(OutputStream os,int n) throws Exception
	{
		os.write(n&0xff);
		os.write(n>>8);
	}

	private Point3D[][] framesDim(float zoom)
	{
		Point3D[][] dims = new Point3D[frames.length][2];
		Point3D[] bb;
		for (int i=0;i<frames.length;i++)
		{
			bb = frames[i].getBoundingBox();
			if (bb[0].x<0)
				bb[0].x=-bb[0].x;
			if (bb[0].y<0)
				bb[0].y=-bb[0].y;
			if (bb[0].z<0)
				bb[0].z=-bb[0].z;
			if (bb[1].x<0)
				bb[1].x=-bb[1].x;
			if (bb[1].y<0)
				bb[1].y=-bb[1].y;
			if (bb[1].z<0)
				bb[1].z=-bb[1].z;
			dims[i][0] = new Point3D();
			dims[i][0].x = Math.max(bb[0].x*zoom,bb[1].x*zoom);
			dims[i][0].y = Math.max(bb[0].y*zoom,bb[1].y*zoom);
			dims[i][0].z = Math.max(bb[0].z*zoom,bb[1].z*zoom);
			dims[i][1] = new Point3D();
			dims[i][1].x = Math.min(bb[0].x*zoom,bb[1].x*zoom);
			dims[i][1].y = bb[0].y*zoom;
			dims[i][1].z = Math.min(bb[0].z*zoom,bb[1].z*zoom);
		}
		return dims;
	}

	public void yetiExport(String fname,float zoom) throws Exception
	{
		JDialog d = new JDialog();
		d.setTitle("Select animations");
		JPanel p = new JPanel(new GridLayout(names.size(),1));
		JCheckBox[] cb = new JCheckBox[names.size()];
		for (int i=0;i<names.size();i++)
		{
			String n = names.get(i);
			p.add(cb[i] = new JCheckBox(n));
			boolean b = true;
			String[] s = new String[] {"stand","run","attack","pain","paina","pain1","death","deatha","death1"};
			for (int j=0;j<s.length && b;j++)
				b = !s[j].equals(n);
			cb[i].setSelected(!b);
		}
		d.getContentPane().setLayout(new BorderLayout());
		d.getContentPane().add(new JScrollPane(p),BorderLayout.CENTER);
		d.setDefaultCloseOperation(JDialog.HIDE_ON_CLOSE);
		d.setModal(true);
		d.setSize(300,200);
		d.setVisible(true);

		System.err.println("Export yeti :"+fname);

		BufferedOutputStream fos = new BufferedOutputStream(new FileOutputStream(fname));

		// get palette
		ArrayList<Integer> pal = new ArrayList<Integer>();
		int[] raster = new int[16384];
		m_skin0.getRGB(0,0,128,128,raster,0,128);
		for (int i=0;i<16384;i++)
		{
			Integer ii = new Integer(raster[i]);
			if (!pal.contains(ii))
				pal.add(ii);
		}
		// color count
		fos.write(pal.size());
		System.err.println(" pal size :"+pal.size());
		// write palette
		for (int i=0;i<pal.size();i++)
		{
			int c = ((Integer)pal.get(i)).intValue();
			fos.write((c>>16)&0xff);
			fos.write((c>>8)&0xff);
			fos.write(c&0xff);
		}
		// write texture
		for (int i=0;i<16384;i++)
			fos.write(pal.indexOf(new Integer(raster[i])));

		for (int i=0;i<export.length;i++)
			export[i] = true;
		
		// count anims
		int k = 0;
		for (int i=0;i<cb.length;i++)
			if (cb[i].isSelected())
				k+=1;
		fos.write(k);

		// frames/anim
		k = 0;
		for (int i=0;i<cb.length;i++)
		{
			if (!cb[i].isSelected())
				disableExport(i);
			else
			{
				System.err.println(" anim "+cb[i].getText()+" = "+frameCount(cb[i].getText()));
				fos.write(k);
				k+=frameCount(cb[i].getText());
			}
		}
		// total frames
		fos.write(k);
		System.err.println(" nb frames :"+k);

		// anims data
		zoom*=10;
		Point3D[][] fdim = framesDim(zoom);
		float[] max = new float[frames.length];
		for (int i=0;i<frames.length;i++)
		{
			if (!export[i])
				continue;
			write16(fos,(int)Math.min(fdim[i][1].x,fdim[i][1].z)); // radius
			max[i] = Math.max(Math.max(fdim[i][0].x,fdim[i][0].z),fdim[i][0].y);
		}
		for (int i=0;i<frames.length;i++)
		{
			if (!export[i])
				continue;
			write16(fos,(int)fdim[i][1].y); // height
		}

		// model data
		int nbt = frames[0].getTriangles().length;
		System.err.println(" nbt :"+nbt);
		write16(fos,nbt);
		fos.write(3); // trianges
		fos.write(0xff); // texture id unknown
		int nbv = frames[0].getVertices().length;
		System.err.println(" nbv :"+nbv);
		write16(fos,nbv);

		// triangles definition
		Triangle3D[] t = frames[0].getTriangles();
		for (int j=0;j<nbt;j++)
		{
			TexturedTriangle3D tj = (TexturedTriangle3D)t[j];
			write16(fos,tj.a);
			fos.write((int)(tj.tp[0].x*m_w0));
			fos.write((int)(tj.tp[0].y*m_h0));
			write16(fos,tj.b);
			fos.write((int)(tj.tp[1].x*m_w0));
			fos.write((int)(tj.tp[1].y*m_h0));
			write16(fos,tj.c);
			fos.write((int)(tj.tp[2].x*m_w0));
			fos.write((int)(tj.tp[2].y*m_h0));
		}
		// animation frames
		for (int i=0;i<frames.length;i++)
		{
			if (!export[i])
				continue;
			float scale = (zoom*127)/max[i];
			fos.write((int)(32*max[i]/127));
			Point3D[] v = frames[i].getVertices();
			for (int j=0;j<nbv;j++)
			{
				fos.write((int)(-v[j].x*scale));
				fos.write((int)(v[j].y*scale));
				fos.write((int)(-v[j].z*scale));
			}
		}
		fos.close();
	}
	
	public void smoothAnim(int n)
	{
		ArrayList<ArrayList<Object>> faces = new ArrayList<ArrayList<Object>>();
		Triangle3D[] tt = frames[0].getTriangles();
		for (int j=0;j<tt.length;j++)
		{
			TexturedTriangle3D t = (TexturedTriangle3D)tt[j];
			ArrayList<Object> v = new ArrayList<Object>();
			v.add(new Integer(t.a));
			v.add(new Integer(t.b));
			v.add(new Integer(t.c));
			ArrayList<Object> att = new ArrayList<Object>();
			att.add(t.texture);
			att.add(new Integer(t.tp[0].x));
			att.add(new Integer(t.tp[0].y));
			att.add(new Integer(t.tp[1].x));
			att.add(new Integer(t.tp[1].y));
			att.add(new Integer(t.tp[2].x));
			att.add(new Integer(t.tp[2].y));
			v.add(att);
			faces.add(v);
		}
		ArrayList<Point3D> vertices = new ArrayList<Point3D>();
		int l = 0;
		DefaultObject3D[] frm = new DefaultObject3D[(frames.length-1)*(n+1)+1];
		int[] anm = new int[frm.length];
		boolean[] exp = new boolean[frm.length];
		int cur = 0;
		for (int i=0;i<frames.length-1;i++)
		{
			frm[l] = frames[i];
			exp[l] = export[i];
			anm[l++] = anims[i];
			Point3D[] p0 = frames[i].getVertices();
			Point3D[] p1;
			if (anims[i+1]==anims[i])
				p1 = frames[i+1].getVertices();
			else
			{
				p1 = frames[cur].getVertices();
				cur = i+1;
			}
			vertices.clear();
			for (int j=1;j<=n;j++)
			{
				for (int k=0;k<p0.length;k++)
				{
					Point3D p = new Point3D();
					p.x = ((p0[k].x+p1[k].x)*j)/(n+1);
					p.y = ((p0[k].y+p1[k].y)*j)/(n+1);
					p.z = ((p0[k].z+p1[k].z)*j)/(n+1);
					vertices.add(p);
				}
				frm[l] = new DefaultObject3D(name+"_"+l,vertices,faces);
				exp[l] = export[i];
				anm[l++] = anims[i];
			}
		}
		frm[l] = frames[frames.length-1];
		exp[l] = export[frames.length-1];
		anm[l++] = anims[frames.length-1];
		frames = frm;
		anims = anm;
		export = exp;
		setFrame(0);
	}
}
