package j3d.loader;

import j3d.Object3D;
import j3d.Point3D;
import j3d.Rotation3D;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.Vector;

public class MS3DObject3DLoader
{
	private String name;
	private byte[] buffer;
	private int offset;

	public MS3DObject3DLoader()
	{
	}

	public Object3D loadObject(String fileName)
	{
		return loadObject(new File(fileName));
	}

	public Object3D loadObject(File file)
	{
		try
		{
			name = file.getAbsolutePath();
			buffer = new byte[(int)file.length()];
			return loadObject(new FileInputStream(file));
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		return null;
	}

	private int readChar()
	{
		if (offset==buffer.length)
			return -1;
		int b = buffer[offset++];
		b= (b<0?256:0)+b;
		return b;
	}

	private int readInt()
	{
		int i = readChar();
		i |= (readChar()<<8);
		return i;
	}

	private int readLong()
	{
		int i = readChar();
		i |= (readChar()<<8);
		i |= (readChar()<<16);
		i |= (readChar()<<24);
		return i;
	}

	private float readFloat()
	{
		int i = readLong();
		return Float.intBitsToFloat(i);
	}
	
	private String readString(int l)
	{
		String s = "";
		char c;
		int off = offset+l;
		for (int i=0;i<l && (c=(char)readChar())!=0;i++)
			s += c;
		offset = off;
		return s;
	}

	public Object3D loadObject(InputStream fos) throws Exception
	{
		fos.read(buffer);
		offset=0;

		Vector<Vector<Object>> faces = new Vector<Vector<Object>>();
		Vector<Point3D> vertices = new Vector<Point3D>();

		String format = readString(10);
		int version = readLong();
		if (!format.equals("MS3D000000")||version<3)
		{
			System.err.println(format+" "+version);
			throw new Exception("Invalid file format");
		}
			
		int nVertices = readInt();
		for (int i = 0; i < nVertices; i++ )
		{
			/*int flag = */readChar();
			Point3D v = new Point3D();
			v.x = readFloat();
			v.y = readFloat();
			v.z = readFloat();
			/*int boneId = */readChar();
			/*int refCount = */readChar();
			vertices.addElement(v);
		}

		int nTriangles = readInt();
		for (int i = 0; i < nTriangles; i++ )
		{
			/*int m_flags = */readInt();
			Vector<Object> face = new Vector<Object>();
			int p0 = readInt();
			int p1 = readInt();
			int p2 = readInt();
			face.addElement(new Integer(p0));
			face.addElement(new Integer(p2));
			face.addElement(new Integer(p1));
			for (int j=0;j<9;j++) // vnorm
				readFloat();
			for (int j=0;j<3;j++) // s?
				readFloat();
			for (int j=0;j<3;j++) // t?
				readFloat();
			/*int smoothingGroup = */readChar();
			/*int groupIndex = */readChar();

			faces.addElement(face);
		}

		int nGroups = readInt();

		for (int i = 0; i < nGroups; i++ )
		{
			readChar();
			readString(32); // name

			int nTrianglesInd = readInt();
			
			int[] pTriangleIndices = new int[nTrianglesInd];						
			for ( int j = 0; j < nTrianglesInd; j++ )
			{
				pTriangleIndices[j] = readInt();
				/*
				(*pTriangles)[pTriangleIndices[j]].mesh=i;
				(*pVertices)[(*pTriangles)[pTriangleIndices[j]].m_vertexIndices[0]].mesh=i;
				(*pVertices)[(*pTriangles)[pTriangleIndices[j]].m_vertexIndices[1]].mesh=i;
				(*pVertices)[(*pTriangles)[pTriangleIndices[j]].m_vertexIndices[2]].mesh=i;
				*/
			}

			/*int materialIndex = */readChar();
/*
			(*pMeshes)[i].m_materialIndex = materialIndex;
			(*pMeshes)[i].m_numTriangles = nTriangles;
			(*pMeshes)[i].m_pTriangleIndices = pTriangleIndices;				
*/
		}

		int nMaterials = readInt();

		for (int i = 0; i < nMaterials; i++ )
		{
			readString(32); // Name
			for (int j=0;j<16;j++) // Lightning
				readFloat();
			readFloat();	// Shininess
			readFloat();	// Transparency
			readChar();		// Mode
			readString(128);// Texture filename
			readString(128);// AlphaMap
		}

		/*float animationFPS = */readFloat();
		/*float currentTime = */readFloat();
		/*int totalFrames = */readLong();

		int nJoints = readInt();

		for (int i = 0; i < nJoints; i++ )
		{
		      readChar();		// Flags
		      readString(32);	// Name
		      readString(32);	// Parent
		      
		      Rotation3D rot = new Rotation3D();
		      rot.x = readFloat();
		      rot.y = readFloat();
		      rot.z = readFloat();
		      Point3D pos = new Point3D();
		      pos.x = readFloat();
		      pos.y = readFloat();
		      pos.z = readFloat();

		      int numKeyFramesRot = readInt();
		      int numKeyFramesPos = readInt();

		      for (int j = 0; j < numKeyFramesRot; j++)
		      {
		    	  readFloat(); // Time
		    	  Rotation3D r = new Rotation3D();
			      r.x = readFloat();
			      r.y = readFloat();
			      r.z = readFloat();
		      }

		      for (int j = 0; j < numKeyFramesPos; j++)
		      {
		    	  readFloat(); // Time
			      Point3D p = new Point3D();
			      p.x = readFloat();
			      p.y = readFloat();
			      p.z = readFloat();
		      }
		}
/*
		for (i=0;i<*numVertices;i++)
		{
			int64 x,y,z;
			x=(*pVertices)[i].m_normal[0];
			y=(*pVertices)[i].m_normal[1];
			z=(*pVertices)[i].m_normal[2];
			x=(x*x+y*y+z*z);
			x=sqrt((float)x)/65536;
			if (x>0)
			{
				(*pVertices)[i].m_normal[0]=(*pVertices)[i].m_normal[0]/x;
				(*pVertices)[i].m_normal[1]=(*pVertices)[i].m_normal[1]/x;
				(*pVertices)[i].m_normal[2]=(*pVertices)[i].m_normal[2]/x;
			}
		}

		
		int32 *cx,*cy,*cz,*cvert;
		cx=(int32*)malloc((*numMeshes)*sizeof(int32));
		cy=(int32*)malloc((*numMeshes)*sizeof(int32));
		cz=(int32*)malloc((*numMeshes)*sizeof(int32));
		cvert=(int32*)malloc((*numMeshes)*sizeof(int32));
		
		for (i=0;i<(*numMeshes);i++)
		{
			cx[i]=cy[i]=cz[i]=cvert[i]=0;	
		}
		
		for (i=0;i<(*numVertices);i++)
		{
			if ((uint32)((*pVertices)[i].mesh)>=(*numMeshes)) (*pVertices)[i].mesh=0;
			else
			{
				cx[(*pVertices)[i].mesh]+=(*pVertices)[i].m_location[0];	
				cy[(*pVertices)[i].mesh]+=(*pVertices)[i].m_location[1];	
				cz[(*pVertices)[i].mesh]+=(*pVertices)[i].m_location[2];	
				cvert[(*pVertices)[i].mesh]++;
			}
		}
		
		for (i=0;i<(*numMeshes);i++)
		{		
			if (cvert[i]>0)
			{
				(*pMeshes)[i].m_center[0]=0*cx[i]/cvert[i];	
				(*pMeshes)[i].m_center[1]=0*cy[i]/cvert[i];	
				(*pMeshes)[i].m_center[2]=0*cz[i]/cvert[i];	
			}
			else
			{
				(*pMeshes)[i].m_center[0]=0;	
				(*pMeshes)[i].m_center[1]=0;	
				(*pMeshes)[i].m_center[2]=0;	
			}
			(*pMeshes)[i].m_position[0]=0;
			(*pMeshes)[i].m_position[1]=0;
			(*pMeshes)[i].m_position[2]=0;
		}
		
		free(cx);
		free(cy);
		free(cz);
		free(cvert);
		
		free(pBuffer);
//	recreer manuellement coord de mapping pour mesh n j	
//	en se basant les coords des points * 1<<k
		int j=0;
		int k;
		for (j=0;j<(*numMeshes);j++)
		{
			k=3;
			if ((model==0)&&(j<3)) k=5;
			for (i=0;i<(*pMeshes)[j].m_numTriangles;i++)
			{

		 		(*pTriangles)[(*pMeshes)[j].m_pTriangleIndices[i]].m_s[0]=(((*pVertices)[(*pTriangles)[(*pMeshes)[j].m_pTriangleIndices[i]].m_vertexIndices[0]].m_location[0])>>(_fixpoint_shifter-k))&TEXTURE_MASK;
	 			(*pTriangles)[(*pMeshes)[j].m_pTriangleIndices[i]].m_s[1]=(((*pVertices)[(*pTriangles)[(*pMeshes)[j].m_pTriangleIndices[i]].m_vertexIndices[1]].m_location[0])>>(_fixpoint_shifter-k))&TEXTURE_MASK;
	 			(*pTriangles)[(*pMeshes)[j].m_pTriangleIndices[i]].m_s[2]=(((*pVertices)[(*pTriangles)[(*pMeshes)[j].m_pTriangleIndices[i]].m_vertexIndices[2]].m_location[0])>>(_fixpoint_shifter-k))&TEXTURE_MASK;
		 		(*pTriangles)[(*pMeshes)[j].m_pTriangleIndices[i]].m_t[0]=(((*pVertices)[(*pTriangles)[(*pMeshes)[j].m_pTriangleIndices[i]].m_vertexIndices[0]].m_location[1])>>(_fixpoint_shifter-k))&TEXTURE_MASK;
		 		(*pTriangles)[(*pMeshes)[j].m_pTriangleIndices[i]].m_t[1]=(((*pVertices)[(*pTriangles)[(*pMeshes)[j].m_pTriangleIndices[i]].m_vertexIndices[1]].m_location[1])>>(_fixpoint_shifter-k))&TEXTURE_MASK;
		 		(*pTriangles)[(*pMeshes)[j].m_pTriangleIndices[i]].m_t[2]=(((*pVertices)[(*pTriangles)[(*pMeshes)[j].m_pTriangleIndices[i]].m_vertexIndices[2]].m_location[1])>>(_fixpoint_shifter-k))&TEXTURE_MASK;
			}
		}
		return 0;
*/
		return new DefaultObject3D(name,vertices,faces);
	}	
}
