package j3d.loader;

import java.awt.*;
import java.io.*;
import java.awt.image.*;
import java.util.*;

public class PCXDecoder {

  public static Image getImage(String fileName) throws IOException
  {
    File f = new File(fileName);
    byte[] buf = new byte[(int)f.length()];
    BufferedInputStream bis = new BufferedInputStream(new FileInputStream(f));
    bis.read(buf);
    bis.close();
    return decode(buf);
  }

  private static int offset;

  private static int btoi(byte b)
  {
    int a = b;
    return (a<0?256+a:a);
  }

  private static int readShort(byte[] buf) {
    int res = btoi(buf[offset]) + (btoi(buf[offset+1])<<8);
    offset+=2;
    return res;
  }

  private static int readByte(byte[] buf) {
    return btoi(buf[offset++]);
  }

  private static int adjust(int c)
  {
	return (c<0xe0)? (c&0xe0) + 0x10 : 0xff;
/*
    if (((c&0x1f)>=0x10))
		return (c<0xe0)? (c&0xe0) + 0x10 : 0xff;
    else
       return c&0xe0;
*/
  }

  private static int makecol(int r,int g,int b)
  {
    return (255<<24)+(adjust(r)<<16)+(adjust(g)<<8)+adjust(b);
  }

  public static Image decode(byte[] buf) throws IOException {
    Vector<Integer> pal = new Vector<Integer>();
	offset = 4;
    int x = readShort(buf),
        y = readShort(buf),
		width = readShort(buf)-x+1,
        height = readShort(buf)-y+1;
	int bpl;
	byte nbp;
	offset+=53;
	nbp = buf[offset++];	// nb planes
	bpl = readShort(buf);	// BytesPerLine
	offset+=60;

    int[] line = new int[bpl*nbp];
    int[] pixels = new int[width*height];
	int i,j,k,l,b;
	for (i=0;i<height;i++)
	{
    	for (j=0;j<bpl*nbp;)
    	{
			b = readByte(buf);
			if ((b&0xc0)==0xc0)
			{
				l = b&0x3f;
				b = readByte(buf);
				for (k=0;k<l;k++)
					line[j++] = b;
			}
			else
				line[j++] = b;
	    }
		for (j=0;j<width;j++)
		{
			if (nbp==1)
			{
				l = (buf.length-768)+line[j]*3;
				pixels[i*width+j] = makecol(buf[l],buf[l+1],buf[l+2]);
			}
			else
				pixels[i*width+j] = makecol(line[j],line[j+bpl],line[j+bpl+bpl]);
			Integer pix = new Integer(pixels[i*width+j]);
			if (!pal.contains(pix))
				pal.addElement(pix);
		}
	}
System.err.println("palsize="+pal.size());
int[] rle = get_rle_sprite(pixels,width,height);
System.err.println(pixels.length+" "+rle.length);
    BufferedImage bimg = new BufferedImage(width,height,BufferedImage.TYPE_INT_ARGB);
    bimg.setRGB(0,0,width,height,rle_sprite(rle,width,height),0,width);
    return bimg;
  }

  private static void gm_memcpy(int[] tdst,int dst,int[] tsrc,int src,int l)
  {
	for (int i=0;i<l;i++)
		tdst[dst++] = tsrc[src++];
  }

  private static int[] get_rle_sprite(int[] b,int w,int h)
  {
	int temp[] = new int[w*h*2];
	int dst = 0;
	int c,s0 = 0,s1 = 0,end = w*h;
	int n;

	n = 0;
	while (s1!=end)
	{
		c = b[s1++];
		n+=1;
		if (s1==end)
			break;
		if (c!=b[s1])
		{
			if (n<127)
				continue;
			temp[dst++] = n;
			gm_memcpy(temp,dst,b,s0,n);
			dst += n;
			n = 0;
			s0 = s1;
		}
		else
		{
			if (n>1)
			{
				n-=1;
				temp[dst++] = n;
				gm_memcpy(temp,dst,b,s0,n);
				dst += n;
			}
			n = 1;
			while ((s1!=end) && (c==b[s1]) && (n<127))
			{
				n+=1;
				s1+=1;
			}
			temp[dst++] = n|0x80;
			temp[dst++] = c;
			n = 0;
			s0 = s1;
		}
	}
	if (n>0)
	{
		temp[dst++] = n;
		gm_memcpy(temp,dst,b,s0,n);
		dst += n;
	}
	int[] buf = new int[dst];
	gm_memcpy(buf,0,temp,0,dst);
	return buf;
  }

  private static int[] rle_sprite(int[] b,int w,int h)
  {
	int temp[] = new int[w*h];
	int i = 0,end = w*h,k=0;
	while (k<end)
	{
		int n = b[i++];
		if ((n&0x80)==0x80)
		{
			n&=0x7f;
			for (int j=0;j<n;j++)
				temp[k++] = b[i];
			i+=1;
		}
		else
		{
			gm_memcpy(temp,k,b,i,n);
			k+=n;
			i+=n;
		}
	}
    return temp;
  }
}
