package j3d.loader;

import java.io.*;
import java.util.*;
import j3d.*;

public class RAWObject3DLoader implements Object3DLoader {

  private String name;

  public RAWObject3DLoader() {
  }

  public Object3D loadObject(String fileName) {
    name = fileName;
    return loadObject(new File(fileName));
  }

  public Object3D loadObject(File file) {
    try {
       name = file.getAbsolutePath();
       return loadObject(new FileInputStream(file));
    } catch(Exception e) { e.printStackTrace(); }
    return null;
  }

  public Object3D loadObject(InputStream is) throws IOException {
      Reader r = new BufferedReader(new InputStreamReader(is));
      StreamTokenizer st = new StreamTokenizer(r);
      st.eolIsSignificant(true);
      st.commentChar('#');
      Vector<Vector<Object>> faces = new Vector<Vector<Object>>();
      Vector<Point3D> vertices = new Vector<Point3D>();
      int numCol = -1;
      DefaultScene3D scene = new DefaultScene3D(name);
      boolean cont = true;
      while (cont) {
          switch (st.nextToken()) {
            case StreamTokenizer.TT_EOL:
              break;
            case StreamTokenizer.TT_EOF:
              cont = false;
            case StreamTokenizer.TT_WORD:
              if (vertices.size()>0) {
                 scene.addObject(new DefaultObject3D(name+String.valueOf(numCol),vertices,faces),new Position3D(new Point3D(0,0,0),new Rotation3D(0,0,0),1f));
              }
              numCol = ((numCol+1)%7)+1;
              vertices.removeAllElements();
              faces.removeAllElements();
              break;
            case StreamTokenizer.TT_NUMBER: {
              double x = 0, y = 0, z = 0;
              // first point
              x = st.nval;
              st.nextToken();
              y = st.nval;
              st.nextToken();
              z = st.nval;
              Point3D p3d = new Point3D((float)x,(float)y,(float)z); 
              int p1 = vertices.indexOf(p3d);
              if (p1==-1) {
                 p1 = vertices.size();
                 vertices.add(p3d);
              }
              // second point
              st.nextToken();
              x = st.nval;
              st.nextToken();
              y = st.nval;
              st.nextToken();
              z = st.nval;
              p3d = new Point3D((float)x,(float)y,(float)z); 
              int p2 = vertices.indexOf(p3d);
              if (p2==-1) {
                 p2 = vertices.size();
                 vertices.add(p3d);
              }
              // last point
              st.nextToken();
              x = st.nval;
              st.nextToken();
              y = st.nval;
              st.nextToken();
              z = st.nval;
              p3d = new Point3D((float)x,(float)y,(float)z); 
              int p3 = vertices.indexOf(p3d);
              if (p3==-1) {
                 p3 = vertices.size();
                 vertices.add(p3d);
              }
              // face
              Vector<Object> face = new Vector<Object>();
              face.add(new Integer(p3));
              face.add(new Integer(p2));
              face.add(new Integer(p1));
              face.add(new Integer(numCol));
              faces.add(face);
              // flush line
              while (st.ttype != StreamTokenizer.TT_EOL) st.nextToken();
            } break;
          }
      }
      is.close();
      return scene;
  }

  public Object3D loadObject(Object source) {
    return null;
  }
}
