package j3d.loader;

import java.io.*;
import java.util.*;
import j3d.*;
import java.awt.Color;
import java.awt.image.BufferedImage;

public class RSDObject3DLoader implements Object3DLoader {

  private String path;
  private String name;
  private Hashtable<String,Color> colors = new Hashtable<String,Color>();
  private Hashtable<String,Vector<Object>> textures = new Hashtable<String,Vector<Object>>();
  private short[][][][] images;

  public RSDObject3DLoader(String path) {
    this.path = path;
  }

  public RSDObject3DLoader() {
    this(null);
  }

  public Object3D loadObject(String fileName) {
    return loadObject(new File(fileName));
  }

  public Object3D loadObject(File file) {
    try {
       String s = name = file.getAbsolutePath();
       setWorkDir(s.substring(0,s.lastIndexOf(File.separator)+1));
       return loadObject(new FileInputStream(file));
    } catch(Exception e) { e.printStackTrace(); }
    return null;
  }

  public void setWorkDir(String path) {
    this.path = path;
  }

  public Object3D loadObject(InputStream is) throws IOException {
    if (path==null)
       throw new IOException();
    LineNumberReader lnr = new LineNumberReader(new InputStreamReader(is));
    String l,ply=null,tx[] = null,mat = null;
    while ((l=lnr.readLine())!=null) {
        l = l.trim();
        if (l.startsWith("PLY=")) {
           ply = l.substring(4);
        } else if (l.startsWith("MAT=")) {
           mat = l.substring(4);
        } else if (l.startsWith("GRP=")) {
        } else if (l.startsWith("NTEX=")) {
           tx = new String[Integer.valueOf(l.substring(5)).intValue()];
        } else if (l.startsWith("TEX[")) {
           tx[Integer.valueOf(l.substring(4,l.indexOf("]"))).intValue()] = l.substring(l.indexOf("=")+1);
        }
    }
    is.close();
    if (tx!=null) {
       images = new short[tx.length][][][];
       for (int i=0;i<tx.length;i++)
           images[i] = TexturedTriangle3D.getTexture(path+tx[i],(BufferedImage)TIMDecoder.getImage(path+tx[i]));
    }
    loadMat(new FileInputStream(new File(path+mat)));
    return loadPly(new FileInputStream(new File(path+ply)));
  }

  public static String readString(InputStream is) throws Exception {
    StringBuffer sb = new StringBuffer();
    int c;
    while ((c=is.read())!=-1) sb.append((char)c);
    return sb.toString();
  }

  private void loadMat(InputStream is) throws IOException {
    LineNumberReader lnr = new LineNumberReader(new InputStreamReader(is));
    String line;
    boolean first=true;
    int nmat = 0, nm = 0;
    while ((line = lnr.readLine())!=null) {
        if (line.startsWith("#")||line.startsWith("@")) continue;
        StringTokenizer st = new StringTokenizer(line);
        if (first) {
           nmat = Integer.valueOf(st.nextElement().toString()).intValue();
           first = false;
           continue;
        }
        if (nm < nmat) {
           String num = st.nextElement().toString();
           /*int flag = */Integer.valueOf(st.nextElement().toString()).intValue();
           /*char shading = */st.nextElement().toString().charAt(0);
           char type = st.nextElement().toString().charAt(0);
           if (type=='C') {
              int r = Integer.valueOf(st.nextElement().toString()).intValue();
              int g = Integer.valueOf(st.nextElement().toString()).intValue();
              int b = Integer.valueOf(st.nextElement().toString()).intValue();
              colors.put(num,new Color(r,g,b));
           } else {
              Vector<Object> v = new Vector<Object>();
              v.addElement(images[Integer.valueOf(st.nextElement().toString()).intValue()]);
              while (st.hasMoreElements())
                  v.addElement(Integer.valueOf(st.nextElement().toString()));
              textures.put(num,v);
           }
           nm++;
        }
    }
  }

  private Object3D loadPly(InputStream is) throws IOException {
    LineNumberReader lnr = new LineNumberReader(new InputStreamReader(is));
    Vector<Vector<Object>> faces = new Vector<Vector<Object>>();
    Vector<Point3D> vertices = new Vector<Point3D>();
    int nv = 0,np = 0, nn = 0,nvert = 0,npoly = 0,nnorm = 0;
    boolean first=true;
    String line;
    while ((line = lnr.readLine())!=null) {
        if (line.startsWith("#")||line.startsWith("@")) continue;
        StringTokenizer st = new StringTokenizer(line);
        if (first) {
           nvert = Integer.valueOf(st.nextElement().toString()).intValue();
           nnorm = Integer.valueOf(st.nextElement().toString()).intValue();
           npoly = Integer.valueOf(st.nextElement().toString()).intValue();
           first = false;
        } else if (nv < nvert) {
           double x = Double.valueOf(st.nextElement().toString()).doubleValue();
           double y = Double.valueOf(st.nextElement().toString()).doubleValue();
           double z = Double.valueOf(st.nextElement().toString()).doubleValue();
           vertices.add(new Point3D((float)x,(float)y,(float)z));
           nv++;
        } else if (nn < nnorm) {
           nn++;
        } else if (np < npoly) {
           int flag = Integer.valueOf(st.nextElement().toString()).intValue();
           int p1 = Integer.valueOf(st.nextElement().toString()).intValue();
           int p2 = Integer.valueOf(st.nextElement().toString()).intValue();
           int p3 = Integer.valueOf(st.nextElement().toString()).intValue();
           int p4 = Integer.valueOf(st.nextElement().toString()).intValue();
           Vector<Object> face = new Vector<Object>();
           face.add(new Integer(p1));
           face.add(new Integer(p2));
           face.add(new Integer(p3));
           Color col = (Color)colors.get(""+np);
           if (col!=null) face.add(col);
           else {
        	  Vector<Object> v = textures.get(""+np);
              Vector<Object> _v = new Vector<Object>(7);
              for (int i=0;i<7;i++) _v.addElement(v.elementAt(i));
              face.add(_v);
           }
           faces.add(face);
           if (flag==1) {
              face = new Vector<Object>();
              face.add(new Integer(p3));
              face.add(new Integer(p2));
              face.add(new Integer(p4));
              col = (Color)colors.get(""+np);
              if (col!=null) face.add(col);
              else {
                 Vector<Object> v = textures.get(""+np);
                 Vector<Object> _v = new Vector<Object>(7);
                 _v.addElement(v.elementAt(0));
                 _v.addElement(v.elementAt(5));
                 _v.addElement(v.elementAt(6));
                 _v.addElement(v.elementAt(3));
                 _v.addElement(v.elementAt(4));
                 _v.addElement(v.elementAt(7));
                 _v.addElement(v.elementAt(8));
                 face.add(_v);
              }
              faces.add(face);
           }
           np++;
       }
    }
    is.close();
    return new DefaultObject3D(name,vertices,faces);
  }

  public Object3D loadObject(Object source) {
    return null;
  }
}
