package j3d.loader;

import java.awt.*;
import java.io.*;
import java.awt.image.*;

public class TIMDecoder {

  public static Image getImage(String fileName) throws IOException {
    File f = new File(fileName);
    byte[] buf = new byte[(int)f.length()];
    BufferedInputStream bis = new BufferedInputStream(new FileInputStream(f));
    bis.read(buf);
    bis.close();
    return decode(buf);
  }

  private static int offset;

  private static int btoi(byte b) {
    int a = b;
    return (a<0?256+a:a);
  }

  private static int readInt(byte[] buf) {
    int res = btoi(buf[offset]) + (btoi(buf[offset+1])<<8) + (btoi(buf[offset+2])<<16) + (btoi(buf[offset+3])<<24);
    offset+=4;
    return res;
  }

  private static int readShort(byte[] buf) {
    int res = btoi(buf[offset]) + (btoi(buf[offset+1])<<8);
    offset+=2;
    return res;
  }

  public static int short2Int(int v) {
    double d = 255.0/31.0;
    int res = ((v&0x8000)==0?255:127)<<24;
    int b = (v&0x7c00)>>10;
    int g = (v&0x3e0)>>5;
    int r = v&0x1f;
    res += (((int)(r*d))<<16)+(((int)(g*d))<<8)+((int)(b*d));
    return res;
  }

  public static Image decode(byte[] buf) throws IOException {
    offset = 0;

    // Reading header
    readInt(buf);
    int numBits = readInt(buf);

    // Color table
    int[] clutdata = null;
    if (numBits!=2) { // No clut for 16bits
       int clutsize = readInt(buf);
       /*int clutw = */readShort(buf);
       /*int cluth = */readShort(buf);
       /*int clutx = */readShort(buf);
       /*int cluty = */readShort(buf);
       clutdata = new int[clutsize-12];
       for (int i=0;i<clutdata.length/2;i++) {
           int v = readShort(buf);
           if (v==0) clutdata[i]=0;
           else if (v==0x8000) clutdata[i]=(255<<24);
           else clutdata[i]=short2Int(v);
       }
    }

    // Pixel data
    /*int pixelsize = */readInt(buf);
    /*int x = */readShort(buf);
    /*int y = */readShort(buf);
    int width = readShort(buf),
        height = readShort(buf);
    int mult = (numBits==8?4:(numBits==9?2:1));
    int[] pixels = new int[width*mult*height];
    int off = 0;

    switch(numBits) {
      case 8 : // 4 bits
        for (int iy = 0;iy < height; iy++) {
            for (int ix = 0;ix < width * mult; ix+=4) {
                int v = readShort(buf);
                for (int i=0;i<4;i++) {
                    pixels[off++] = clutdata[v&0xf];
                    v>>=4;
                }
            }
        }
        break;
      case 9 : // 8 bits
        for (int iy = 0;iy < height; iy++) {
            for (int ix = 0;ix < width * mult; ix+=2) {
                int v = readShort(buf);
                pixels[off++] = clutdata[v&0xff];
                pixels[off++] = clutdata[v>>8];
            }
        }
        break;
      case 2 : // 16 bits
        for (int iy = 0;iy < height; iy++) {
            for (int ix = 0;ix < width * mult; ix++) {
                pixels[off++] = short2Int(readShort(buf));
            }
        }
        break;
    }
    BufferedImage bimg = new BufferedImage(width*mult,height,BufferedImage.TYPE_INT_ARGB);
    bimg.setRGB(0,0,width*mult,height,pixels,0,width*mult);
    return bimg;
  }
}
